"""
CrossFormer, adapted from
https://github.com/cheerss/CrossFormer/blob/crossformer/models/crossformer.py

Paper "CrossFormer: A Versatile Vision Transformer Hinging on Cross-scale Attention",
https://arxiv.org/abs/2108.00154

Changes from original:
* Dynamic group size based on the input size
"""

# Reference license: MIT

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import Permute
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class PatchEmbed(nn.Module):
    def __init__(self, patch_sizes: list[int], in_channels: int, embed_dim: int) -> None:
        super().__init__()
        self.embed_dim = embed_dim

        self.projs = nn.ModuleList()
        for i, ps in enumerate(patch_sizes):
            if i == len(patch_sizes) - 1:
                dim = embed_dim // 2**i
            else:
                dim = embed_dim // 2 ** (i + 1)

            stride = patch_sizes[0]
            padding = (ps - patch_sizes[0]) // 2
            self.projs.append(
                nn.Conv2d(in_channels, dim, kernel_size=(ps, ps), stride=(stride, stride), padding=(padding, padding))
            )

        self.norm = nn.LayerNorm(embed_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        xs = []
        for proj in self.projs:
            xs.append(proj(x).flatten(2).transpose(1, 2))  # B Ph*Pw C

        x = torch.concat(xs, dim=2)
        x = self.norm(x)

        return x


class DynamicPosBias(nn.Module):
    def __init__(self, dim: int, num_heads: int) -> None:
        super().__init__()
        pos_dim = dim // 4
        self.pos = nn.Sequential(
            nn.Linear(2, pos_dim),
            nn.LayerNorm(pos_dim),
            nn.ReLU(),
            nn.Linear(pos_dim, pos_dim),
            nn.LayerNorm(pos_dim),
            nn.ReLU(),
            nn.Linear(pos_dim, pos_dim),
            nn.LayerNorm(pos_dim),
            nn.ReLU(),
            nn.Linear(pos_dim, num_heads),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.pos(x)


class Attention(nn.Module):
    def __init__(
        self, dim: int, group_size: tuple[int, int], num_heads: int, qkv_bias: bool, attn_drop: float, proj_drop: float
    ) -> None:
        super().__init__()
        self.group_size = group_size  # Wh, Ww
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.pos = DynamicPosBias(dim // 4, self.num_heads)

        self.define_bias_table()
        self.define_relative_position_index()

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def define_bias_table(self) -> None:
        position_bias_h = torch.arange(1 - self.group_size[0], self.group_size[0])
        position_bias_w = torch.arange(1 - self.group_size[1], self.group_size[1])
        biases = torch.stack(torch.meshgrid([position_bias_h, position_bias_w], indexing="ij"))  # 2, 2Wh-1, 2W2-1
        biases = biases.flatten(1).transpose(0, 1).float()
        self.biases = nn.Buffer(biases)

    def define_relative_position_index(self) -> None:
        coords_h = torch.arange(self.group_size[0])
        coords_w = torch.arange(self.group_size[1])
        coords = torch.stack(torch.meshgrid([coords_h, coords_w], indexing="ij"))  # 2, Wh, Ww
        coords_flatten = torch.flatten(coords, 1)  # 2, Wh*Ww
        relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]  # 2, Wh*Ww, Wh*Ww
        relative_coords = relative_coords.permute(1, 2, 0).contiguous()  # Wh*Ww, Wh*Ww, 2
        relative_coords[:, :, 0] += self.group_size[0] - 1  # shift to start from 0
        relative_coords[:, :, 1] += self.group_size[1] - 1
        relative_coords[:, :, 0] *= 2 * self.group_size[1] - 1
        relative_position_index = relative_coords.sum(-1)  # Wh*Ww, Wh*Ww
        self.relative_position_index = nn.Buffer(relative_position_index)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, N, C) = x.size()
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        (q, k, v) = qkv.unbind(0)

        q = q * self.scale
        attn = q @ k.transpose(-2, -1)

        pos = self.pos(self.biases)  # 2Wh-1 * 2Ww-1, heads
        relative_position_bias = pos[self.relative_position_index.view(-1)].view(
            self.group_size[0] * self.group_size[1], self.group_size[0] * self.group_size[1], -1
        )  # Wh*Ww,Wh*Ww,nH
        relative_position_bias = relative_position_bias.permute(2, 0, 1).contiguous()  # nH, Wh*Ww, Wh*Ww
        attn = attn + relative_position_bias.unsqueeze(0)
        attn = attn.softmax(dim=-1)

        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class CrossFormerBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        input_resolution: tuple[int, int],
        num_heads: int,
        group_size: tuple[int, int],
        use_lda: bool,
        mlp_ratio: float,
        qkv_bias: bool,
        proj_drop: float,
        attn_drop: float,
        drop_path: float,
    ) -> None:
        super().__init__()
        self.dim = dim
        self.input_resolution = input_resolution
        self.num_heads = num_heads
        self.group_size = group_size
        self.use_lda = use_lda
        self.mlp_ratio = mlp_ratio
        if self.input_resolution[0] <= self.group_size[0]:
            self.use_lda = False
            self.group_size = (self.input_resolution[0], self.group_size[1])
        if self.input_resolution[1] <= self.group_size[1]:
            self.use_lda = False
            self.group_size = (self.group_size[0], self.input_resolution[1])

        self.norm1 = nn.LayerNorm(dim)
        self.attn = Attention(
            dim,
            group_size=self.group_size,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            attn_drop=attn_drop,
            proj_drop=proj_drop,
        )

        self.norm2 = nn.LayerNorm(dim)
        self.mlp = MLP(dim, [int(dim * mlp_ratio), dim], activation_layer=nn.GELU, dropout=proj_drop)

        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (H, W) = self.input_resolution
        (B, _, C) = x.size()

        shortcut = x
        x = self.norm1(x)
        x = x.view(B, H, W, C)

        # Group embeddings
        (GH, GW) = self.group_size  # pylint: disable=invalid-name
        if self.use_lda is False:
            x = x.reshape(B, H // GH, GH, W // GW, GW, C).permute(0, 1, 3, 2, 4, 5)
        else:
            x = x.reshape(B, GH, H // GH, GW, W // GW, C).permute(0, 2, 4, 1, 3, 5)

        x = x.reshape(B * H * W // (GH * GW), GH * GW, C)

        # Attention
        x = self.attn(x)  # nW*B, G*G, C

        # Un-group embeddings
        x = x.reshape(B, H // GH, W // GW, GH, GW, C)
        if self.use_lda is False:
            x = x.permute(0, 1, 3, 2, 4, 5).reshape(B, H, W, C)
        else:
            x = x.permute(0, 3, 1, 4, 2, 5).reshape(B, H, W, C)

        x = x.view(B, H * W, C)

        # FFN
        x = shortcut + self.drop_path(x)
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x


class PatchMerging(nn.Module):
    def __init__(self, input_resolution: tuple[int, int], dim: int, patch_sizes: list[int]) -> None:
        super().__init__()
        self.input_resolution = input_resolution
        self.dim = dim
        self.reductions = nn.ModuleList()
        self.norm = nn.LayerNorm(dim)

        for i, ps in enumerate(patch_sizes):
            if i == len(patch_sizes) - 1:
                out_dim = 2 * dim // 2**i
            else:
                out_dim = 2 * dim // 2 ** (i + 1)

            stride = 2
            padding = (ps - stride) // 2
            self.reductions.append(
                nn.Conv2d(dim, out_dim, kernel_size=(ps, ps), stride=(stride, stride), padding=(padding, padding))
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (H, W) = self.input_resolution
        (B, _, C) = x.shape

        x = self.norm(x)
        x = x.view(B, H, W, C).permute(0, 3, 1, 2)

        xs = []
        for reduction in self.reductions:
            xs.append(reduction(x).flatten(2).transpose(1, 2))

        x = torch.concat(xs, dim=2)

        return x


class CrossFormerStage(nn.Module):
    def __init__(
        self,
        prev_dim: int,
        dim: int,
        input_resolution: tuple[int, int],
        depth: int,
        num_heads: int,
        group_size: tuple[int, int],
        mlp_ratio: float,
        qkv_bias: bool,
        proj_drop: float,
        attn_drop: float,
        drop_path: list[float],
        downsample: bool,
        patch_sizes: list[int],
    ) -> None:
        super().__init__()
        if downsample is True:
            self.downsample = PatchMerging(input_resolution, dim=prev_dim, patch_sizes=patch_sizes)
            input_resolution = (input_resolution[0] // 2, input_resolution[1] // 2)
        else:
            self.downsample = nn.Identity()

        layers = []
        for i in range(depth):
            use_lda = (i % 2) != 0
            layers.append(
                CrossFormerBlock(
                    dim=dim,
                    input_resolution=input_resolution,
                    num_heads=num_heads,
                    group_size=group_size,
                    use_lda=use_lda,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    proj_drop=proj_drop,
                    attn_drop=attn_drop,
                    drop_path=drop_path[i],
                )
            )

        self.blocks = nn.Sequential(*layers)
        self.resolution = input_resolution

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)

        return x


class CrossFormer(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        patch_sizes = [4, 8, 16, 32]
        merge_sizes = [[2, 4], [2, 4], [2, 4]]
        embed_dim: int = self.config["embed_dim"]
        depths: list[int] = self.config["depths"]
        num_heads: list[int] = self.config["num_heads"]
        drop_path_rate: float = self.config["drop_path_rate"]
        group_size = (int(self.size[0] / (2**5)), int(self.size[1] / (2**5)))

        self.patch_sizes = patch_sizes
        self.patch_embed = PatchEmbed(patch_sizes=patch_sizes, in_channels=self.input_channels, embed_dim=embed_dim)
        patch_resolution = (self.size[0] // patch_sizes[0], self.size[1] // patch_sizes[0])

        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]
        num_stages = len(depths)
        prev_dim = embed_dim
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_stages):
            patch_sizes = merge_sizes[i - 1] if i > 0 else [0]
            dim = int(embed_dim * 2**i)
            if i == 0:
                input_resolution = patch_resolution
            else:
                input_resolution = (patch_resolution[0] // (2 ** (i - 1)), patch_resolution[1] // (2 ** (i - 1)))

            stages[f"stage{i+1}"] = CrossFormerStage(
                prev_dim=prev_dim,
                dim=dim,
                input_resolution=input_resolution,
                depth=depths[i],
                num_heads=num_heads[i],
                group_size=group_size,
                mlp_ratio=4.0,
                qkv_bias=True,
                proj_drop=0.0,
                attn_drop=0.0,
                drop_path=dpr[i],
                downsample=i > 0,
                patch_sizes=patch_sizes,
            )
            return_channels.append(dim)
            prev_dim = dim

        last_features = int(embed_dim * 2 ** (num_stages - 1))
        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.LayerNorm(last_features),
            Permute([0, 2, 1]),
            nn.AdaptiveAvgPool1d(output_size=1),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = last_features
        self.classifier = self.create_classifier()

        # Weights initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.LayerNorm):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.patch_embed(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                (H, W) = module.resolution
                (B, _, C) = x.size()
                out[name] = x.view(B, H, W, C).permute(0, 3, 1, 2).contiguous()

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.patch_embed.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.patch_embed(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        super().adjust_size(new_size)

        new_patch_resolution = (new_size[0] // self.patch_sizes[0], new_size[1] // self.patch_sizes[0])
        input_resolution = new_patch_resolution
        for mod in self.body.modules():
            if isinstance(mod, CrossFormerStage):
                for m in mod.modules():
                    if isinstance(m, PatchMerging):
                        m.input_resolution = input_resolution
                        input_resolution = (input_resolution[0] // 2, input_resolution[1] // 2)
                    elif isinstance(m, CrossFormerBlock):
                        m.input_resolution = input_resolution

                mod.resolution = input_resolution

        new_group_size = (int(new_size[0] / (2**5)), int(new_size[1] / (2**5)))
        for m in self.body.modules():
            if isinstance(m, CrossFormerBlock):
                m.group_size = new_group_size
                if m.input_resolution[0] <= m.group_size[0]:
                    m.use_lda = False
                    m.group_size = (m.input_resolution[0], m.group_size[1])
                if m.input_resolution[1] <= m.group_size[1]:
                    m.use_lda = False
                    m.group_size = (m.group_size[0], m.input_resolution[1])

            elif isinstance(m, Attention):
                m.group_size = new_group_size
                m.define_bias_table()
                m.define_relative_position_index()


registry.register_model_config(
    "crossformer_t",
    CrossFormer,
    config={"embed_dim": 64, "depths": [1, 1, 8, 6], "num_heads": [2, 4, 8, 16], "drop_path_rate": 0.1},
)
registry.register_model_config(
    "crossformer_s",
    CrossFormer,
    config={"embed_dim": 96, "depths": [2, 2, 6, 2], "num_heads": [3, 6, 12, 24], "drop_path_rate": 0.2},
)
registry.register_model_config(
    "crossformer_b",
    CrossFormer,
    config={"embed_dim": 96, "depths": [2, 2, 18, 2], "num_heads": [3, 6, 12, 24], "drop_path_rate": 0.3},
)
registry.register_model_config(
    "crossformer_l",
    CrossFormer,
    config={"embed_dim": 128, "depths": [2, 2, 18, 2], "num_heads": [4, 8, 16, 32], "drop_path_rate": 0.5},
)

registry.register_weights(
    "crossformer_s_arabian-peninsula256px",
    {
        "url": "https://huggingface.co/birder-project/crossformer_s_arabian-peninsula/resolve/main",
        "description": "CrossFormer small model trained on the arabian-peninsula dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 116.7,
                "sha256": "24feadef6e7921702fade34487bfcf65f0656151e2a00ede6a8ae719534d68c6",
            }
        },
        "net": {"network": "crossformer_s", "tag": "arabian-peninsula256px"},
    },
)
registry.register_weights(
    "crossformer_s_arabian-peninsula",
    {
        "url": "https://huggingface.co/birder-project/crossformer_s_arabian-peninsula/resolve/main",
        "description": "CrossFormer small model trained on the arabian-peninsula dataset",
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 118.3,
                "sha256": "94936dfb3f641cbd227b90ada937eb5055e176fa95f517711e14315b2e0b9405",
            }
        },
        "net": {"network": "crossformer_s", "tag": "arabian-peninsula"},
    },
)



import logging
import os
from collections.abc import Callable
from pathlib import Path
from typing import Any
from typing import Literal
from typing import Optional

import torch
import webdataset as wds
from torchvision.datasets.folder import IMG_EXTENSIONS
from torchvision.io import ImageReadMode
from torchvision.io import decode_image

from birder.common import fs_ops
from birder.common.training_utils import reduce_across_processes
from birder.conf import settings

logger = logging.getLogger(__name__)

WDS_SHUFFLE_SIZE = int(os.environ.get("WDS_SHUFFLE_SIZE", 4000))
WDS_INITIAL_SIZE = int(os.environ.get("WDS_INITIAL_SIZE", 1000))


def decode_sample_name(item: tuple[str, str, Any, int]) -> tuple[str, Any, int]:
    sample_name = item[0] + "/" + item[1]
    return (sample_name, item[2], item[3])


def decode_fake_cls(item: tuple[Any, ...]) -> tuple[Any, ...]:
    return (*item, settings.NO_LABEL)


def wds_image_decoder(key: str, data: bytes) -> torch.Tensor:
    if key.endswith(IMG_EXTENSIONS) is False:
        return None

    tensor = torch.frombuffer(bytearray(data), dtype=torch.uint8)
    return decode_image(tensor, mode=ImageReadMode.RGB)


def identity(x: Any) -> Any:
    return x


def make_wds_dataset(
    wds_path: str | list[str],
    dataset_size: int,
    shuffle: bool,
    samples_names: bool,
    transform: Callable[..., torch.Tensor],
    img_loader: Literal["tv", "pil"] = "tv",
    *,
    cls_key: Optional[str] = "cls",
    cache_dir: Optional[str] = None,
    shuffle_buffer_size: Optional[int] = None,
    shuffle_initial_size: Optional[int] = None,
) -> torch.utils.data.IterableDataset:
    if shuffle is True:
        shardshuffle = 500
    else:
        shardshuffle = False

    dataset = wds.WebDataset(
        wds_path, shardshuffle=shardshuffle, nodesplitter=wds.split_by_node, cache_dir=cache_dir, empty_check=False
    )
    if shuffle is True:
        if shuffle_buffer_size is None:
            shuffle_buffer_size = WDS_SHUFFLE_SIZE
        if shuffle_initial_size is None:
            shuffle_initial_size = WDS_INITIAL_SIZE

        logger.debug(f"Using buffer size of {shuffle_buffer_size} for shuffle with {shuffle_initial_size} initial size")
        dataset = dataset.shuffle(shuffle_buffer_size, initial=shuffle_initial_size)

    return_keys = ["jpeg;jpg;png;webp"]
    if cls_key is not None:
        return_keys = return_keys + [cls_key]
    if samples_names is True:
        return_keys = ["__url__", "__key__"] + return_keys

    if img_loader == "pil":
        dataset = dataset.with_length(dataset_size, silent=True).decode("pil").to_tuple(*return_keys)
    else:
        dataset = dataset.with_length(dataset_size, silent=True).decode(wds_image_decoder).to_tuple(*return_keys)

    if cls_key is None:
        dataset = dataset.map(decode_fake_cls)

    if samples_names is True:
        dataset = dataset.map(decode_sample_name)
        dataset = dataset.map_tuple(identity, transform, identity)
    else:
        dataset = dataset.map_tuple(transform, identity)

    return dataset


def wds_size(wds_path: str, device: torch.device) -> int:
    dataset = wds.WebDataset(
        wds_path,
        shardshuffle=False,
        select_files=lambda key_name: key_name.endswith("cls"),
        nodesplitter=wds.split_by_node,
        empty_check=False,
    ).batched(64, collation_fn=None, partial=True)
    dataloader = wds.WebLoader(dataset, batch_size=None, num_workers=8)
    size = 0
    for batch in dataloader:
        size += len(batch)

    size = reduce_across_processes(size, device, op=torch.distributed.ReduceOp.SUM)  # type: ignore

    return size


def prepare_wds_args(data_path: str, size: Optional[int], device: torch.device) -> tuple[str, int]:
    if "://" not in data_path:
        if ".." not in data_path:
            # Local path without braces, build brace argument
            (wds_path, _) = fs_ops.wds_braces_from_path(Path(data_path))

        else:
            wds_path = data_path

        if size is None:
            # If size not provided, we scan all tar files
            dataset_size = wds_size(wds_path, device)
        else:
            dataset_size = size

    else:
        # Remote path, we take the string as-is
        # Dataset size and must be provided
        assert size is not None
        wds_path = data_path
        dataset_size = size

    return (wds_path, dataset_size)


def wds_args_from_info(info_path: str, split: str) -> tuple[list[str], int]:
    info = fs_ops.read_wds_info(info_path)
    root = Path(info_path).parent

    size = info["splits"][split]["num_samples"]
    filenames = info["splits"][split]["filenames"]

    filenames = [os.path.join(root, f) for f in filenames]

    return (filenames, size)



import logging
import tarfile
from pathlib import Path

from birder.common import cli

logger = logging.getLogger(__name__)


def download_url(url: str, target: str | Path, sha256: str, progress_bar: bool = True) -> bool:
    if isinstance(target, str):
        target = Path(target)

    if target.exists() is True:
        if cli.calc_sha256(target) == sha256:
            logger.debug("File already downloaded and verified")
            return False

        raise RuntimeError("Downloaded file is corrupted")

    target.parent.mkdir(parents=True, exist_ok=True)
    logger.info(f"Downloading {url} to {target}")
    cli.download_file(url, target, sha256, progress_bar=progress_bar)
    return True


def extract_archive(from_path: str | Path, to_path: str | Path) -> None:
    logger.info(f"Extracting {from_path} to {to_path}")
    with tarfile.open(from_path, "r") as tar:
        if hasattr(tarfile, "data_filter") is True:
            tar.extractall(to_path, filter="data")
        else:
            # NOTE: Remove once minimum Python version is 3.12 or above
            tar.extractall(to_path)  # nosec # tarfile_unsafe_members



"""
I-JEPA (Image-based Joint-Embedding Predictive Architecture), adapted from
https://github.com/facebookresearch/ijepa/blob/main/src/models/vision_transformer.py

Paper "Self-Supervised Learning from Images with a Joint-Embedding Predictive Architecture",
https://arxiv.org/abs/2301.08243

Changes from original:
* Removed weight initialization with scaling per layer index
"""

# Reference license: Attribution-NonCommercial 4.0 International

import math
from typing import Any
from typing import Optional

import torch
from torch import nn

from birder.net.base import MaskedTokenOmissionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import pos_embedding_sin_cos_2d
from birder.net.ssl.base import SSLBaseNet
from birder.net.vit import Encoder


def apply_masks(x: torch.Tensor, masks: list[torch.Tensor]) -> torch.Tensor:
    all_x = []
    for m in masks:
        mask_keep = m.unsqueeze(-1).repeat(1, 1, x.size(-1))
        all_x.append(torch.gather(x, dim=1, index=mask_keep))

    return torch.concat(all_x, dim=0)


def repeat_interleave_batch(x: torch.Tensor, b: int, repeat: int) -> torch.Tensor:
    N = len(x) // b
    x = torch.concat([torch.concat([x[i * b : (i + 1) * b] for _ in range(repeat)], dim=0) for i in range(N)], dim=0)
    return x


class MultiBlockMasking:
    # Adapted from: https://github.com/facebookresearch/ijepa/blob/main/src/masks/multiblock.py

    def __init__(
        self,
        input_size: tuple[int, int],
        enc_mask_scale: tuple[float, float],
        pred_mask_scale: tuple[float, float],
        aspect_ratio: tuple[float, float],
        n_enc: int,
        n_pred: int,
        min_keep: int,
        allow_overlap: bool,
    ):
        self.height = input_size[0]
        self.width = input_size[1]
        self.enc_mask_scale = enc_mask_scale
        self.pred_mask_scale = pred_mask_scale
        self.aspect_ratio = aspect_ratio
        self.n_enc = n_enc
        self.n_pred = n_pred
        self.min_keep = min_keep  # Minimum number of patches to keep
        self.allow_overlap = allow_overlap  # Whether to allow overlap between enc and pred masks

    def _sample_block_size(
        self, scale: tuple[float, float], aspect_ratio_scale: tuple[float, float]
    ) -> tuple[int, int]:
        _rand = torch.rand(1).item()

        (min_s, max_s) = scale
        mask_scale = min_s + _rand * (max_s - min_s)
        max_keep = int(self.height * self.width * mask_scale)

        (min_ar, max_ar) = aspect_ratio_scale
        aspect_ratio = min_ar + _rand * (max_ar - min_ar)

        # Compute block height and width (given scale and aspect-ratio)
        h = int(round(math.sqrt(max_keep * aspect_ratio)))
        w = int(round(math.sqrt(max_keep / aspect_ratio)))
        while h >= self.height:
            h -= 1
        while w >= self.width:
            w -= 1

        return (h, w)

    def _sample_block_mask(
        self, b_size: tuple[int, int], acceptable_regions: Optional[list[torch.Tensor]] = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        h, w = b_size

        def constrain_mask(mask: torch.Tensor, tries: int = 0) -> None:
            """
            Helper to restrict given mask to a set of acceptable regions
            """

            assert acceptable_regions is not None
            N = max(int(len(acceptable_regions) - tries), 0)
            for k in range(N):
                mask *= acceptable_regions[k]

        # Loop to sample masks until we find a valid one
        tries = 0
        timeout = 20
        og_timeout = 20
        valid_mask = False
        while valid_mask is False:
            # Sample block top-left corner
            top = torch.randint(0, self.height - h, (1,))
            left = torch.randint(0, self.width - w, (1,))
            mask = torch.zeros((self.height, self.width), dtype=torch.int32)
            mask[top : top + h, left : left + w] = 1

            # Constrain mask to a set of acceptable regions
            if acceptable_regions is not None:
                constrain_mask(mask, tries)

            mask = torch.nonzero(mask.flatten())

            # If mask too small try again
            valid_mask = len(mask) > self.min_keep
            if valid_mask is False:
                timeout -= 1
                if timeout == 0:
                    tries += 1
                    timeout = og_timeout

        mask = mask.squeeze()
        mask_complement = torch.ones((self.height, self.width), dtype=torch.int32)
        mask_complement[top : top + h, left : left + w] = 0

        return (mask, mask_complement)

    def __call__(self, batch_size: int) -> tuple[list[torch.Tensor], list[torch.Tensor]]:
        """
        Create encoder and predictor masks
        1. sample enc block (size + location)
        2. sample pred block (size)
        3. sample several enc block locations for each image
        4. sample several pred block locations for each image
        5. return enc mask and pred mask
        """

        p_size = self._sample_block_size(scale=self.pred_mask_scale, aspect_ratio_scale=self.aspect_ratio)
        e_size = self._sample_block_size(scale=self.enc_mask_scale, aspect_ratio_scale=(1.0, 1.0))

        collated_masks_pred = []
        collated_masks_enc = []
        min_keep_pred = self.height * self.width
        min_keep_enc = self.height * self.width
        for _ in range(batch_size):
            masks_p = []
            masks_c = []
            for _ in range(self.n_pred):
                (mask, mask_c) = self._sample_block_mask(p_size)
                masks_p.append(mask)
                masks_c.append(mask_c)
                min_keep_pred = min(min_keep_pred, len(mask))

            collated_masks_pred.append(masks_p)

            acceptable_regions: Optional[list[torch.Tensor]] = masks_c
            if self.allow_overlap is True:
                acceptable_regions = None

            masks_e = []
            for _ in range(self.n_enc):
                (mask, _) = self._sample_block_mask(e_size, acceptable_regions=acceptable_regions)
                masks_e.append(mask)
                min_keep_enc = min(min_keep_enc, len(mask))

            collated_masks_enc.append(masks_e)

        collated_masks_pred = [[cm[:min_keep_pred] for cm in cm_list] for cm_list in collated_masks_pred]
        collated_masks_pred = torch.utils.data.default_collate(collated_masks_pred)

        collated_masks_enc = [[cm[:min_keep_enc] for cm in cm_list] for cm_list in collated_masks_enc]
        collated_masks_enc = torch.utils.data.default_collate(collated_masks_enc)

        return (collated_masks_enc, collated_masks_pred)


class VisionTransformerPredictor(nn.Module):
    def __init__(
        self,
        size: tuple[int, int],
        embed_dim: int,
        predictor_embed_dim: int,
        mlp_dim: int,
        num_heads: int,
        depth: int,
        drop_path_rate: float,
    ) -> None:
        super().__init__()
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]

        self.predictor_embed = nn.Linear(embed_dim, predictor_embed_dim)
        self.mask_token = nn.Parameter(torch.zeros(1, 1, predictor_embed_dim))

        pos_embedding = pos_embedding_sin_cos_2d(h=size[0], w=size[1], dim=predictor_embed_dim, num_special_tokens=0)
        self.pos_embedding = nn.Parameter(pos_embedding, requires_grad=False)

        self.encoder = Encoder(
            depth, num_heads, predictor_embed_dim, mlp_dim, dropout=0.0, attention_dropout=0.0, dpr=dpr
        )
        self.norm = nn.LayerNorm(predictor_embed_dim, eps=1e-6)
        self.predictor_proj = nn.Linear(predictor_embed_dim, embed_dim)

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.LayerNorm):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

            elif isinstance(m, nn.Conv2d):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def forward(
        self, x: torch.Tensor, masks_encoder: list[torch.Tensor], masks_pred: list[torch.Tensor]
    ) -> torch.Tensor:
        B = len(x) // len(masks_encoder)

        # Project to predictor dim
        x = self.predictor_embed(x)

        # Apply positional embedding
        x_pos_embed = self.pos_embedding.repeat(B, 1, 1)
        x = x + apply_masks(x_pos_embed, masks_encoder)

        # Add positional embedding to mask tokens
        ctx = x.size(1)
        pos_embed = self.pos_embedding.repeat(B, 1, 1)
        pos_embed = apply_masks(pos_embed, masks_pred)
        pos_embed = repeat_interleave_batch(pos_embed, B, repeat=len(masks_encoder))

        pred_tokens = self.mask_token.repeat(pos_embed.size(0), pos_embed.size(1), 1)
        pred_tokens = pred_tokens + pos_embed

        # Concat mask tokens
        x = x.repeat(len(masks_pred), 1, 1)
        x = torch.concat([x, pred_tokens], dim=1)

        # Encoder forward
        x = self.encoder(x)
        x = self.norm(x)

        # Mask tokens predictions
        x = x[:, ctx:]
        x = self.predictor_proj(x)

        return x


# pylint: disable=invalid-name
class I_JEPA(SSLBaseNet):
    def __init__(
        self,
        backbone: PreTrainEncoder,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(backbone, config=config, size=size)
        assert self.config is None, "config not supported"
        assert isinstance(self.backbone, MaskedTokenOmissionMixin)

    def forward(self, x: torch.Tensor, ids_keep: Optional[torch.Tensor] = None) -> torch.Tensor:
        return self.backbone.masked_encoding_omission(x, ids_keep)["tokens"]



"""
Xception, adapted from
https://github.com/keras-team/keras/blob/r2.15/keras/applications/xception.py

Paper "Xception: Deep Learning with Depthwise Separable Convolutions", https://arxiv.org/abs/1610.02357
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation

from birder.net.base import DetectorBackbone


class SeparableConv2d(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
    ) -> None:
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels,
            in_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            groups=in_channels,
            bias=False,
        )
        self.pointwise = nn.Conv2d(
            in_channels, out_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        x = self.pointwise(x)
        return x


class XceptionBlock(nn.Module):
    def __init__(
        self, in_channels: int, out_channels: int, repeats: int, stride: tuple[int, int], grow_first: bool
    ) -> None:
        super().__init__()

        if out_channels != in_channels or stride[0] != 1 or stride[1] != 1:
            self.skip = Conv2dNormActivation(
                in_channels,
                out_channels,
                kernel_size=(1, 1),
                stride=stride,
                padding=(0, 0),
                bias=False,
                activation_layer=None,
            )

        else:
            self.skip = nn.Identity()

        layers = []
        for i in range(repeats):
            if grow_first is True:
                out_c = out_channels
                if i == 0:
                    in_c = in_channels
                else:
                    in_c = out_channels

            else:
                in_c = in_channels
                if i < (repeats - 1):
                    out_c = in_channels
                else:
                    out_c = out_channels

            layers.append(nn.ReLU(inplace=True))
            layers.append(SeparableConv2d(in_c, out_c, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1)))
            layers.append(nn.BatchNorm2d(out_c))

        if stride[0] != 1 or stride[1] != 1:
            layers.append(nn.MaxPool2d(kernel_size=(3, 3), stride=stride, padding=(1, 1)))

        self.block = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        branch = self.block(x)
        identity = self.skip(x)
        x = branch + identity

        return x


class Xception(DetectorBackbone):
    default_size = (299, 299)
    auto_register = True

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is None, "config not supported"

        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels, 32, kernel_size=(3, 3), stride=(2, 2), padding=(0, 0), bias=False
            ),
            Conv2dNormActivation(
                32, 64, kernel_size=(3, 3), stride=(1, 1), padding=(0, 0), bias=False, activation_layer=None
            ),  # Remove ReLU here, first Xception block starts with ReLU
        )

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []

        stages["stage1"] = XceptionBlock(64, 128, repeats=2, stride=(2, 2), grow_first=True)
        return_channels.append(128)

        stages["stage2"] = XceptionBlock(128, 256, repeats=2, stride=(2, 2), grow_first=True)
        return_channels.append(256)

        stages["stage3"] = nn.Sequential(
            XceptionBlock(256, 728, repeats=2, stride=(2, 2), grow_first=True),
            XceptionBlock(728, 728, repeats=3, stride=(1, 1), grow_first=True),
            XceptionBlock(728, 728, repeats=3, stride=(1, 1), grow_first=True),
            XceptionBlock(728, 728, repeats=3, stride=(1, 1), grow_first=True),
            XceptionBlock(728, 728, repeats=3, stride=(1, 1), grow_first=True),
            XceptionBlock(728, 728, repeats=3, stride=(1, 1), grow_first=True),
            XceptionBlock(728, 728, repeats=3, stride=(1, 1), grow_first=True),
            XceptionBlock(728, 728, repeats=3, stride=(1, 1), grow_first=True),
            XceptionBlock(728, 728, repeats=3, stride=(1, 1), grow_first=True),
        )
        return_channels.append(728)

        stages["stage4"] = nn.Sequential(
            XceptionBlock(728, 1024, repeats=2, stride=(2, 2), grow_first=False),
            SeparableConv2d(1024, 1536, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1)),
            nn.BatchNorm2d(1536),
            nn.ReLU(inplace=True),
            SeparableConv2d(1536, 2048, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1)),
            nn.BatchNorm2d(2048),
            nn.ReLU(inplace=True),
        )
        return_channels.append(2048)

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = 2048
        self.classifier = self.create_classifier()

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")

            elif isinstance(m, nn.BatchNorm2d):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)



"""
Projected Gradient Descent, adapted from
https://github.com/Harry24k/adversarial-attacks-pytorch/blob/master/torchattacks/attacks/pgd.py

Paper "Towards Deep Learning Models Resistant to Adversarial Attacks",
https://arxiv.org/abs/1706.06083
"""

# Reference license: MIT

from typing import NamedTuple
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn

PGDResponse = NamedTuple("PGDResponse", [("out", torch.Tensor), ("adv_img", torch.Tensor), ("adv_out", torch.Tensor)])


class PGD:
    def __init__(self, net: nn.Module, eps: float, max_delta: float, steps: int, random_start: bool) -> None:
        self.net = net
        self.max_delta = max_delta
        self.eps = eps
        self.steps = steps
        self.random_start = random_start

    def __call__(self, input_tensor: torch.Tensor, target: Optional[torch.Tensor]) -> PGDResponse:
        adv_image = input_tensor.clone().detach()
        out = self.net(input_tensor)
        if target is None:
            target = torch.argmax(out, dim=1)

        if self.random_start is True:
            # Starting at a uniformly random point
            adv_image = adv_image + torch.empty_like(adv_image).uniform_(-self.max_delta, self.max_delta)
            adv_image = torch.clamp(adv_image, min=-4, max=4).detach()

        for _ in range(self.steps):
            adv_image.requires_grad = True
            outputs = self.net(adv_image)
            loss = F.nll_loss(outputs, target)
            self.net.zero_grad()
            loss.backward()

            grad = adv_image.grad.data
            adv_image = adv_image.detach() + self.eps * grad.sign()
            delta = torch.clamp(adv_image - input_tensor, min=-self.max_delta, max=self.max_delta)
            adv_image = torch.clamp(input_tensor + delta, min=-4, max=4).detach()

        adv_out = self.net(adv_image)

        return PGDResponse(F.softmax(out, dim=1), adv_image, F.softmax(adv_out, dim=1))



import argparse
import logging
from typing import Any

from birder.common import cli
from birder.common.lib import get_pretrained_model_url
from birder.conf import settings
from birder.model_registry import registry

logger = logging.getLogger(__name__)


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "download-model",
        allow_abbrev=False,
        help="download pretrained model",
        description="download pretrained model",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools download-model mobilenet_v3_large_1_0\n"
            "python -m birder.tools download-model convnext_v2_tiny_0 --force\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument(
        "--format",
        type=str,
        choices=["pt", "pt2", "pts", "ptl", "safetensors"],
        default="pt",
        help="model serialization format",
    )
    subparser.add_argument("--no-verify", action="store_true", help="skip SHA256 checksum verification")
    subparser.add_argument("--force", action="store_true", help="force download even if model already exists")
    subparser.add_argument("model_name", help="the model to download")
    subparser.set_defaults(func=main)


def main(args: argparse.Namespace) -> None:
    if args.model_name not in registry.list_pretrained_models():
        raise cli.ValidationError(f"{args.model_name} not found, see list-models tool for available options")

    if settings.MODELS_DIR.exists() is False:
        logger.info(f"Creating {settings.MODELS_DIR} directory...")
        settings.MODELS_DIR.mkdir(parents=True)

    model_metadata = registry.get_pretrained_metadata(args.model_name)
    if args.format not in model_metadata["formats"]:
        logger.warning(
            f"Requested format '{args.format}' not found, "
            f"available formats for {args.model_name} are: {list(model_metadata['formats'].keys())}"
        )
        raise SystemExit(1)

    (model_file, url) = get_pretrained_model_url(args.model_name, args.format)
    dst = settings.MODELS_DIR.joinpath(model_file)
    if dst.exists() is True and args.force is False:
        logger.warning(f"File {model_file} already exists... aborting")
        raise SystemExit(1)

    if args.no_verify is True:
        checksum = None
    else:
        checksum = model_metadata["formats"][args.format]["sha256"]

    cli.download_file(url, dst, checksum, override=args.force)



"""
HieraDet, adapted from
https://github.com/facebookresearch/sam2/blob/main/sam2/modeling/backbones/hieradet.py

Paper "SAM 2: Segment Anything in Images and Videos", https://arxiv.org/abs/2408.00714

Changes from original:
* Support only 2d
* Allow dynamic window_spec (by defining divisor as a negative number)
"""

# Reference license: Apache-2.0

import copy
from collections import OrderedDict
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import Permute
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType


def window_partition(x: torch.Tensor, window_size: int) -> tuple[torch.Tensor, tuple[int, int]]:
    (B, H, W, C) = x.size()

    pad_h = (window_size - H % window_size) % window_size
    pad_w = (window_size - W % window_size) % window_size
    if pad_h > 0 or pad_w > 0:
        x = F.pad(x, (0, 0, 0, pad_w, 0, pad_h))

    h_p = H + pad_h
    w_p = W + pad_w

    x = x.view(B, h_p // window_size, window_size, w_p // window_size, window_size, C)
    windows = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(-1, window_size, window_size, C)

    return (windows, (h_p, w_p))


def window_unpartition(
    windows: torch.Tensor, window_size: int, pad_hw: tuple[int, int], hw: tuple[int, int]
) -> torch.Tensor:
    (h_p, w_p) = pad_hw
    (H, W) = hw
    B = windows.shape[0] // (h_p * w_p // window_size // window_size)
    x = windows.view(B, h_p // window_size, w_p // window_size, window_size, window_size, -1)
    x = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(B, h_p, w_p, -1)

    if h_p > H or w_p > W:
        x = x[:, :H, :W, :].contiguous()

    return x


class PatchEmbed(nn.Module):
    def __init__(
        self, dim_in: int, dim_out: int, kernel: tuple[int, int], stride: tuple[int, int], padding: tuple[int, int]
    ) -> None:
        super().__init__()
        self.proj = nn.Conv2d(dim_in, dim_out, kernel_size=kernel, stride=stride, padding=padding)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.proj(x)
        x = x.permute(0, 2, 3, 1)

        return x


class MultiScaleAttention(nn.Module):
    def __init__(self, dim: int, dim_out: int, num_heads: int, q_pool: Optional[nn.Module]) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.q_pool = q_pool
        self.qkv = nn.Linear(dim, dim_out * 3)
        self.proj = nn.Linear(dim_out, dim_out)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, H, W, _) = x.size()
        qkv = self.qkv(x).reshape(B, H * W, 3, self.num_heads, -1)
        (q, k, v) = qkv.unbind(dim=2)

        # Q pooling (for downsample at stage changes)
        if self.q_pool is not None:
            q = q.reshape(B, H, W, -1).permute(0, 3, 1, 2)
            q = self.q_pool(q).permute(0, 2, 3, 1)
            (H, W) = q.shape[1:3]  # Downsampled shape
            q = q.reshape(B, H * W, self.num_heads, -1)

        x = F.scaled_dot_product_attention(  # pylint: disable=not-callable
            q.transpose(1, 2),
            k.transpose(1, 2),
            v.transpose(1, 2),
        )
        x = x.transpose(1, 2)
        x = x.reshape(B, H, W, -1)
        x = self.proj(x)

        return x


class MultiScaleBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        dim_out: int,
        num_heads: int,
        mlp_ratio: float,
        drop_path: float,
        q_stride: Optional[tuple[int, int]],
        window_size: int,
    ) -> None:
        super().__init__()
        self.dim = dim
        self.dim_out = dim_out

        self.norm1 = nn.LayerNorm(dim)
        if dim != dim_out:
            self.proj = nn.Linear(dim, dim_out)
        else:
            self.proj = nn.Identity()

        self.window_size = window_size
        self.q_stride = q_stride
        if self.q_stride is not None:
            self.pool = nn.MaxPool2d(kernel_size=q_stride, stride=q_stride, ceil_mode=False)
        else:
            self.pool = None

        self.attn = MultiScaleAttention(
            dim,
            dim_out,
            num_heads=num_heads,
            q_pool=copy.deepcopy(self.pool),
        )
        self.norm2 = nn.LayerNorm(dim_out)
        self.mlp = MLP(dim_out, [int(dim_out * mlp_ratio), dim_out], activation_layer=nn.GELU)
        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        shortcut = x  # (B, H, W, C)
        x = self.norm1(x)

        # Skip connection
        if self.dim != self.dim_out:
            shortcut = self.proj(x)
            if self.pool is not None:
                shortcut = shortcut.permute(0, 3, 1, 2)
                shortcut = self.pool(shortcut).permute(0, 2, 3, 1)

        # Window partition
        window_size = self.window_size
        pad_hw = (shortcut.size(1), shortcut.size(2))
        (H, W) = pad_hw
        if self.window_size > 0:
            H = x.shape[1]
            W = x.shape[2]
            (x, pad_hw) = window_partition(x, window_size)

        # Window Attention + Q Pooling (if stage change)
        x = self.attn(x)
        if self.q_stride is not None:
            # Shapes have changed due to Q pooling
            window_size = self.window_size // self.q_stride[0]
            (H, W) = (shortcut.size(1), shortcut.size(2))

            pad_h = (window_size - H % window_size) % window_size
            pad_w = (window_size - W % window_size) % window_size
            pad_hw = (H + pad_h, W + pad_w)

        # Reverse window partition
        if self.window_size > 0:
            x = window_unpartition(x, window_size, pad_hw, (H, W))

        x = shortcut + self.drop_path(x)
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x


class HieraDet(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.stage(\d+)\.(\d+)"

    # pylint: disable=too-many-locals
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        patch_kernel = (7, 7)
        patch_stride = (4, 4)
        patch_padding = (3, 3)
        q_stride = (2, 2)
        q_pool = 3
        dim_mul = 2.0
        head_mul = 2.0
        depths: list[int] = self.config["depths"]
        embed_dim: int = self.config["embed_dim"]
        num_heads: int = self.config["num_heads"]
        global_pos_size: tuple[int, int] = self.config["global_pos_size"]
        global_att_blocks: list[int] = self.config["global_att_blocks"]
        window_spec: list[int] = self.config["window_spec"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.stage_ends = [sum(depths[:i]) - 1 for i in range(1, len(depths) + 1)]
        self.q_pool_blocks = [x + 1 for x in self.stage_ends[:-1]][:q_pool]
        assert 0 <= q_pool <= len(self.stage_ends[:-1])

        stem_dim = embed_dim
        self.stem = PatchEmbed(self.input_channels, stem_dim, patch_kernel, patch_stride, patch_padding)

        self.pos_embed = nn.Parameter(torch.zeros(1, embed_dim, *global_pos_size))
        self.pos_embed_win = nn.Parameter(torch.zeros(1, embed_dim, window_spec[0], window_spec[0]))

        depth = sum(depths)
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  # Stochastic depth decay rule

        cur_stage = 1
        layers = []
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(depth):
            dim_out = embed_dim
            window_size = window_spec[cur_stage - 1]  # Lags by a block

            if i in global_att_blocks:
                window_size = 0
            elif window_size < 0:
                window_size = min(self.size) // -window_size

            if i - 1 in self.stage_ends:
                dim_out = int(embed_dim * dim_mul)
                num_heads = int(num_heads * head_mul)
                cur_stage += 1

            block = MultiScaleBlock(
                dim=embed_dim,
                dim_out=dim_out,
                num_heads=num_heads,
                mlp_ratio=4.0,
                drop_path=dpr[i],
                q_stride=q_stride if i in self.q_pool_blocks else None,
                window_size=window_size,
            )
            embed_dim = dim_out
            layers.append(block)

            if i in self.stage_ends:
                stages[f"stage{cur_stage}"] = nn.Sequential(*layers)
                return_channels.append(dim_out)
                layers = []

        assert len(layers) == 0

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.LayerNorm(embed_dim),
            Permute([0, 3, 1, 2]),  # B H W C -> B C H W
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = embed_dim
        self.classifier = self.create_classifier()

        self.stem_stride = patch_stride[0]
        self.stem_width = stem_dim
        self.encoding_size = embed_dim

        # Weight initialization
        nn.init.trunc_normal_(self.pos_embed, std=0.02)
        nn.init.trunc_normal_(self.pos_embed_win, std=0.02)

    def _get_pos_embed(self, x: torch.Tensor) -> torch.Tensor:
        h = x.size(1)
        w = x.size(2)
        window_embed = self.pos_embed_win
        pos_embed = F.interpolate(self.pos_embed, size=(h, w), mode="bicubic")
        pos_embed = pos_embed + window_embed.tile([i // j for i, j in zip(pos_embed.shape, window_embed.shape)])
        pos_embed = pos_embed.permute(0, 2, 3, 1)

        return pos_embed

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)
        x = x + self._get_pos_embed(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x.permute(0, 3, 1, 2).contiguous()

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(
            x, mask, channels_last=True, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token
        )
        x = x + self._get_pos_embed(x)
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x.permute(0, 3, 1, 2).contiguous()
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        x = x + self._get_pos_embed(x)
        x = self.body(x)

        return x

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        x = self.features(x)

        return x

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        super().adjust_size(new_size)

        assert self.config is not None, "must set config"
        global_att_blocks: list[int] = self.config["global_att_blocks"]
        window_spec: list[int] = self.config["window_spec"]

        cur_stage = 1
        i = 0
        for m in self.body.modules():
            if isinstance(m, MultiScaleBlock):
                window_size = window_spec[cur_stage - 1]
                if i in global_att_blocks:
                    window_size = 0
                elif window_size < 0:
                    window_size = min(new_size) // -window_size

                if i - 1 in self.stage_ends:
                    cur_stage += 1

                m.window_size = window_size

                i += 1

    def load_hiera_weights(self, state_dict: dict[str, Any]) -> None:
        # NOTE: Only abswin variant supported

        # Remove classifier weights
        if "classifier.weight" in state_dict:
            del state_dict["classifier.weight"]
            del state_dict["classifier.bias"]

        # Adjust pos_embed_win
        if self.pos_embed.size(3) != state_dict["pos_embed"].size(3):
            pos_embed = state_dict["pos_embed"]
            pos_embed = F.interpolate(pos_embed, size=(self.pos_embed.shape[2:4]), mode="bicubic")
            state_dict["pos_embed"] = pos_embed

        # Load the modified state dict
        incompatible_keys = self.load_state_dict(state_dict, strict=False)
        assert len(incompatible_keys.unexpected_keys) == 0


registry.register_model_config(
    "hieradet_tiny",
    HieraDet,
    config={
        "depths": [1, 2, 7, 2],
        "embed_dim": 96,
        "num_heads": 1,
        "global_pos_size": (7, 7),
        "global_att_blocks": [5, 7, 9],
        "window_spec": [8, 4, -16, -32],
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hieradet_small",
    HieraDet,
    config={
        "depths": [1, 2, 11, 2],
        "embed_dim": 96,
        "num_heads": 1,
        "global_pos_size": (7, 7),
        "global_att_blocks": [7, 10, 13],
        "window_spec": [8, 4, -16, -32],
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hieradet_base",
    HieraDet,
    config={
        "depths": [2, 3, 16, 3],
        "embed_dim": 96,
        "num_heads": 1,
        "global_pos_size": (14, 14),
        "global_att_blocks": [12, 16, 20],
        "window_spec": [8, 4, -16, -32],
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hieradet_base_plus",
    HieraDet,
    config={
        "depths": [2, 3, 16, 3],
        "embed_dim": 112,
        "num_heads": 2,
        "global_pos_size": (14, 14),
        "global_att_blocks": [12, 16, 20],
        "window_spec": [8, 4, -16, -32],
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hieradet_large",
    HieraDet,
    config={
        "depths": [2, 6, 36, 4],
        "embed_dim": 144,
        "num_heads": 2,
        "global_pos_size": (7, 7),
        "global_att_blocks": [23, 33, 43],
        "window_spec": [8, 4, -16, -32],
        "drop_path_rate": 0.2,
    },
)

registry.register_weights(
    "hieradet_small_dino-v2",
    {
        "url": "https://huggingface.co/birder-project/hieradet_small_dino-v2/resolve/main",
        "description": (
            "HieraDet small image encoder pre-trained using DINOv2. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 129.6,
                "sha256": "eb41b8a35445e7f350797094d5e365306b29351e64edd4a316420c23d1e17073",
            }
        },
        "net": {"network": "hieradet_small", "tag": "dino-v2"},
    },
)



import torch
import torch.nn.functional as F
from torch import nn

from birder.layers.ffn import FFN


class MultiHeadAttentionPool(nn.Module):
    """
    Adapted from:
    https://github.com/huggingface/pytorch-image-models/blob/main/timm/layers/attention_pool.py#L12
    """

    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_dim: int,
        qkv_bias: bool,
        latent_len: int = 1,
    ) -> None:
        super().__init__()
        assert dim % num_heads == 0
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.scale = self.head_dim**-0.5

        self.latent_len = latent_len
        self.latent = nn.Parameter(torch.zeros(1, self.latent_len, dim))

        self.q = nn.Linear(dim, dim, bias=qkv_bias)
        self.kv = nn.Linear(dim, dim * 2, bias=qkv_bias)
        self.proj = nn.Linear(dim, dim)

        self.norm = nn.LayerNorm(dim)
        self.mlp = FFN(dim, mlp_dim, act_layer=nn.GELU)

        # Weight initialization
        nn.init.trunc_normal_(self.latent, std=dim**-0.5)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, N, C) = x.size()

        q_latent = self.latent.expand(B, self.latent_len, -1)
        q = self.q(q_latent).reshape(B, self.latent_len, self.num_heads, self.head_dim).transpose(1, 2)

        kv = self.kv(x).reshape(B, N, 2, self.num_heads, self.head_dim).permute(2, 0, 3, 1, 4)
        (k, v) = kv.unbind(0)

        x = F.scaled_dot_product_attention(q, k, v, scale=self.scale)  # pylint: disable=not-callable
        x = x.transpose(1, 2).reshape(B, self.latent_len, C)
        x = self.proj(x)
        x = x + self.mlp(self.norm(x))

        return x



"""
VICReg, adapted from
https://github.com/facebookresearch/vicreg/blob/main/main_vicreg.py

Paper "VICReg: Variance-Invariance-Covariance Regularization for Self-Supervised Learning",
https://arxiv.org/abs/2105.04906
"""

# Reference license: MIT

from typing import Any
from typing import Optional

import torch
import torch.distributed as dist
import torch.nn.functional as F
from torch import nn

from birder.common import training_utils
from birder.net.base import BaseNet
from birder.net.ssl.base import SSLBaseNet


def off_diagonal(x: torch.Tensor) -> torch.Tensor:
    n = x.size(0)
    # assert x.size(0) == x.size(1)
    return x.flatten()[:-1].view(n - 1, n + 1)[:, 1:].flatten()


# pylint: disable=abstract-method,arguments-differ
class FullGatherLayer(torch.autograd.Function):
    @staticmethod
    def forward(_ctx, x):  # type: ignore
        if training_utils.is_dist_available_and_initialized() is False:
            return (x,)

        output = [torch.zeros_like(x) for _ in range(dist.get_world_size())]
        dist.all_gather(output, x)
        return tuple(output)

    @staticmethod
    def backward(_ctx, *grads):  # type: ignore
        all_gradients = torch.stack(grads)
        if training_utils.is_dist_available_and_initialized() is False:
            return all_gradients

        dist.all_reduce(all_gradients)
        return all_gradients[dist.get_rank()]


class VICReg(SSLBaseNet):
    def __init__(
        self,
        backbone: BaseNet,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(backbone, config=config, size=size)
        assert self.config is not None, "must set config"

        mlp_dim: int = self.config["mlp_dim"]
        batch_size: int = self.config["batch_size"]
        sim_coeff: float = self.config["sim_coeff"]
        std_coeff: float = self.config["std_coeff"]
        cov_coeff: float = self.config["cov_coeff"]
        sync_batches: bool = self.config.get("sync_batches", False)

        self.num_features = mlp_dim
        self.world_batch_size = batch_size * training_utils.get_world_size()
        self.sim_coeff = sim_coeff
        self.std_coeff = std_coeff
        self.cov_coeff = cov_coeff
        self.sync_batches = sync_batches

        self.projector = nn.Sequential(
            nn.Linear(self.backbone.embedding_size, mlp_dim),
            nn.BatchNorm1d(mlp_dim),
            nn.ReLU(),
            nn.Linear(mlp_dim, mlp_dim),
            nn.BatchNorm1d(mlp_dim),
            nn.ReLU(),
            nn.Linear(mlp_dim, mlp_dim, bias=False),
        )

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:  # type: ignore[override]
        x = self.projector(self.backbone.embedding(x))
        y = self.projector(self.backbone.embedding(y))

        repr_loss = F.mse_loss(x, y)

        if self.sync_batches is True:
            x = torch.concat(FullGatherLayer.apply(x), dim=0)
            y = torch.concat(FullGatherLayer.apply(y), dim=0)

        x = x - x.mean(dim=0)
        y = y - y.mean(dim=0)

        std_x = torch.sqrt(x.var(dim=0) + 0.0001)
        std_y = torch.sqrt(y.var(dim=0) + 0.0001)
        std_loss = torch.mean(F.relu(1 - std_x)) / 2 + torch.mean(F.relu(1 - std_y)) / 2

        cov_x = (x.T @ x) / (self.world_batch_size - 1)
        cov_y = (y.T @ y) / (self.world_batch_size - 1)
        cov_loss = off_diagonal(cov_x).pow_(2).sum().div(self.num_features) + off_diagonal(cov_y).pow_(2).sum().div(
            self.num_features
        )

        loss = self.sim_coeff * repr_loss + self.std_coeff * std_loss + self.cov_coeff * cov_loss

        return loss



"""
iFormer (Inception Transformer), adapted from
https://github.com/sail-sg/iFormer/blob/main/models/inception_transformer.py

Paper "Inception Transformer", https://arxiv.org/abs/2205.12956

Changes from original:
* Removed biases before norms
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import Permute
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.layers import LayerScale
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType


class PatchEmbed(nn.Module):
    def __init__(
        self,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
        in_channels: int,
        embed_dim: int,
    ) -> None:
        super().__init__()
        self.proj = nn.Conv2d(
            in_channels, embed_dim, kernel_size=kernel_size, stride=stride, padding=padding, bias=False
        )
        self.norm = nn.BatchNorm2d(embed_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.proj(x)
        x = self.norm(x)
        x = x.permute(0, 2, 3, 1)

        return x


class HighMixer(nn.Module):
    def __init__(
        self, dim: int, kernel_size: tuple[int, int], stride: tuple[int, int], padding: tuple[int, int]
    ) -> None:
        super().__init__()
        self.cnn_in = dim // 2
        cnn_dim = self.cnn_in * 2

        self.conv = nn.Conv2d(self.cnn_in, cnn_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False)
        self.proj1 = nn.Conv2d(
            cnn_dim, cnn_dim, kernel_size=kernel_size, stride=stride, padding=padding, groups=cnn_dim, bias=False
        )
        self.act1 = nn.GELU()

        self.max_pool = nn.MaxPool2d(kernel_size, stride=stride, padding=padding)
        self.proj2 = nn.Conv2d(self.cnn_in, cnn_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.act2 = nn.GELU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        cx = x[:, : self.cnn_in, :, :].contiguous()
        cx = self.conv(cx)
        cx = self.proj1(cx)
        cx = self.act1(cx)

        px = x[:, self.cnn_in :, :, :].contiguous()
        px = self.max_pool(px)
        px = self.proj2(px)
        px = self.act2(px)

        return torch.concat([cx, px], dim=1)


class LowMixer(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        qkv_bias: bool,
        pool_size: tuple[int, int],
        attn_drop: float,
    ) -> None:
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5
        self.dim = dim
        self.attn_drop = attn_drop

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        if pool_size[0] > 1 or pool_size[1] > 1:
            self.pool = nn.AvgPool2d(pool_size, stride=pool_size, padding=0, count_include_pad=False)
            self.upsample = nn.Upsample(scale_factor=pool_size)
        else:
            self.pool = nn.Identity()
            self.upsample = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.pool(x)
        (B, _, H, W) = x.size()
        x = x.permute(0, 2, 3, 1).view(B, -1, self.dim)

        (B, N, C) = x.size()
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        (q, k, v) = qkv.unbind(0)
        x = F.scaled_dot_product_attention(  # pylint: disable=not-callable
            q, k, v, dropout_p=self.attn_drop if self.training else 0.0, scale=self.scale
        )
        x = x.transpose(2, 3).reshape(B, C, N)
        x = x.view(B, C, H, W)
        x = self.upsample(x)

        return x


class Mixer(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        qkv_bias: bool,
        attention_head: int,
        pool_size: tuple[int, int],
        attn_drop: float,
        proj_drop: float,
    ) -> None:
        super().__init__()
        head_dim = dim // num_heads
        low_dim = attention_head * head_dim
        high_dim = dim - low_dim
        self.high_dim = high_dim

        self.high_mixer = HighMixer(high_dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.low_mixer = LowMixer(
            low_dim,
            num_heads=attention_head,
            qkv_bias=qkv_bias,
            pool_size=pool_size,
            attn_drop=attn_drop,
        )

        self.conv_fuse = nn.Conv2d(
            low_dim + high_dim * 2,
            low_dim + high_dim * 2,
            kernel_size=(3, 3),
            stride=(1, 1),
            padding=(1, 1),
            groups=low_dim + high_dim * 2,
            bias=False,
        )
        self.proj = nn.Conv2d(low_dim + high_dim * 2, dim, kernel_size=1, stride=1, padding=0)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.permute(0, 3, 1, 2)

        hx = x[:, : self.high_dim, :, :].contiguous()
        hx = self.high_mixer(hx)

        lx = x[:, self.high_dim :, :, :].contiguous()
        lx = self.low_mixer(lx)

        x = torch.concat([hx, lx], dim=1)
        x = x + self.conv_fuse(x)
        x = self.proj(x)
        x = self.proj_drop(x)
        x = x.permute(0, 2, 3, 1).contiguous()

        return x


class InceptionTransformerBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_ratio: float,
        qkv_bias: bool,
        attention_head: int,
        pool_size: tuple[int, int],
        drop: float,
        attn_drop: float,
        drop_path: float,
        layer_scale_init_value: Optional[float],
    ) -> None:
        super().__init__()
        self.norm1 = nn.LayerNorm(dim, eps=1e-6)
        self.attn = Mixer(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            attention_head=attention_head,
            pool_size=pool_size,
            attn_drop=attn_drop,
            proj_drop=0.0,
        )
        self.drop_path = StochasticDepth(drop_path, mode="row")

        self.norm2 = nn.LayerNorm(dim, eps=1e-6)
        self.mlp = MLP(dim, [int(dim * mlp_ratio), dim], activation_layer=nn.GELU, dropout=drop)

        if layer_scale_init_value is not None:
            self.layer_scale_1 = LayerScale(dim, layer_scale_init_value)
            self.layer_scale_2 = LayerScale(dim, layer_scale_init_value)
        else:
            self.layer_scale_1 = nn.Identity()
            self.layer_scale_2 = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.drop_path(self.layer_scale_1(self.attn(self.norm1(x))))
        x = x + self.drop_path(self.layer_scale_2(self.mlp(self.norm2(x))))

        return x


class InceptionTransformerStage(nn.Module):
    def __init__(
        self,
        dim: int,
        out_dim: int,
        num_heads: int,
        mlp_ratio: float,
        qkv_bias: bool,
        attention_heads: list[int],
        pool_size: tuple[int, int],
        drop: float,
        attn_drop: float,
        drop_path: list[float],
        layer_scale_init_value: Optional[float],
        depth: int,
        resolution: tuple[int, int],
        downsample: bool,
    ) -> None:
        super().__init__()
        self.dynamic_size = False
        self.resolution = resolution
        if downsample is True:
            self.downsample = PatchEmbed(
                kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), in_channels=dim, embed_dim=out_dim
            )
        else:
            assert dim == out_dim
            self.downsample = nn.Identity()

        self.pos_embed = nn.Parameter(torch.zeros(1, resolution[0], resolution[1], out_dim))
        layers = []
        for i in range(depth):
            layers.append(
                InceptionTransformerBlock(
                    out_dim,
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    attention_head=attention_heads[i],
                    pool_size=pool_size,
                    drop=drop,
                    attn_drop=attn_drop,
                    drop_path=drop_path[i],
                    layer_scale_init_value=layer_scale_init_value,
                )
            )

        self.blocks = nn.Sequential(*layers)

        # Weight initialization
        nn.init.trunc_normal_(self.pos_embed, std=0.02)

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        self.dynamic_size = dynamic_size

    def _get_pos_embed(self, H: int, W: int) -> torch.Tensor:
        if self.dynamic_size is False:
            return self.pos_embed

        if H == self.resolution[0] and W == self.resolution[1]:
            return self.pos_embed

        orig_dtype = self.pos_embed.dtype
        pos_embedding = self.pos_embed.float()
        pos_embedding = F.interpolate(pos_embedding.permute(0, 3, 1, 2), size=(H, W), mode="bilinear").permute(
            0, 2, 3, 1
        )
        pos_embedding = pos_embedding.to(orig_dtype)
        return pos_embedding

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        (H, W) = x.shape[1:3]

        x = x + self._get_pos_embed(H, W)
        x = self.blocks(x)
        x = x.permute(0, 3, 1, 2)

        return x


# pylint: disable=invalid-name
class iFormer(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.stage(\d+)\.blocks\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        img_size = self.size
        depths: list[int] = self.config["depths"]
        embed_dims: list[int] = self.config["embed_dims"]
        num_heads: list[int] = self.config["num_heads"]
        attention_heads: list[int] = self.config["attention_heads"]
        layer_scale_init_value: Optional[float] = self.config["layer_scale_init_value"]
        drop_path_rate: float = self.config["drop_path_rate"]

        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]
        num_stages = len(depths)

        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels,
                embed_dims[0] // 2,
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                activation_layer=nn.GELU,
                inplace=None,
            ),
            Conv2dNormActivation(
                embed_dims[0] // 2,
                embed_dims[0],
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                activation_layer=None,
            ),
            Permute([0, 2, 3, 1]),
        )

        head_index = 0
        resolution = (img_size[0] // 4, img_size[1] // 4)
        prev_dim = embed_dims[0]
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_stages):
            stages[f"stage{i+1}"] = InceptionTransformerStage(
                prev_dim,
                embed_dims[i],
                num_heads=num_heads[i],
                mlp_ratio=4.0,
                qkv_bias=True,
                attention_heads=attention_heads[head_index : head_index + depths[i]],
                pool_size=(2, 2) if i < 2 else (1, 1),
                drop=0.0,
                attn_drop=0.0,
                drop_path=dpr[i],
                layer_scale_init_value=layer_scale_init_value if i >= 2 else None,
                depth=depths[i],
                resolution=resolution,
                downsample=i > 0,
            )
            return_channels.append(embed_dims[i])
            prev_dim = embed_dims[i]
            resolution = (resolution[0] // 2, resolution[1] // 2)
            head_index = head_index + depths[i]

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            Permute([0, 2, 3, 1]),  # B C H W -> B H W C
            nn.LayerNorm(embed_dims[-1], eps=1e-6),
            Permute([0, 3, 1, 2]),  # B H W C -> B C H W
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = embed_dims[-1]
        self.classifier = self.create_classifier()

        self.stem_stride = 4
        self.stem_width = embed_dims[0]
        self.encoding_size = embed_dims[-1]

        # Weight initialization
        for m in self.modules():
            if isinstance(m, (nn.Conv2d, nn.Linear)):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, (nn.LayerNorm, nn.GroupNorm, nn.BatchNorm2d)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(
            x, mask, channels_last=True, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token
        )
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        super().set_dynamic_size(dynamic_size)
        for stage in self.body.modules():
            if isinstance(stage, InceptionTransformerStage):
                stage.set_dynamic_size(dynamic_size)

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        super().adjust_size(new_size)

        resolution = (new_size[0] // 4, new_size[1] // 4)
        for stage in self.body.modules():
            if isinstance(stage, InceptionTransformerStage):
                orig_dtype = stage.pos_embed.dtype
                pos_embedding = stage.pos_embed.float()
                pos_embedding = F.interpolate(
                    pos_embedding.permute(0, 3, 1, 2), size=resolution, mode="bilinear"
                ).permute(0, 2, 3, 1)
                pos_embedding = pos_embedding.to(orig_dtype)
                stage.pos_embed = nn.Parameter(pos_embedding)
                stage.resolution = resolution
                resolution = (resolution[0] // 2, resolution[1] // 2)


registry.register_model_config(
    "iformer_s",
    iFormer,
    config={
        "depths": [3, 3, 9, 3],
        "embed_dims": [96, 192, 320, 384],
        "num_heads": [3, 6, 10, 12],
        "attention_heads": [1] * 3 + [3] * 3 + [7] * 4 + [9] * 5 + [11] * 3,
        "layer_scale_init_value": 1e-6,
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "iformer_b",
    iFormer,
    config={
        "depths": [4, 6, 14, 6],
        "embed_dims": [96, 192, 384, 512],
        "num_heads": [3, 6, 12, 16],
        "attention_heads": [1] * 4 + [3] * 6 + [8] * 7 + [10] * 7 + [15] * 6,
        "layer_scale_init_value": 1e-6,
        "drop_path_rate": 0.4,
    },
)
registry.register_model_config(
    "iformer_l",
    iFormer,
    config={
        "depths": [4, 6, 18, 8],
        "embed_dims": [96, 192, 448, 640],
        "num_heads": [3, 6, 14, 20],
        "attention_heads": [1] * 4 + [3] * 6 + [10] * 9 + [12] * 9 + [19] * 8,
        "layer_scale_init_value": 1e-6,
        "drop_path_rate": 0.5,
    },
)

registry.register_weights(
    "iformer_s_arabian-peninsula",
    {
        "url": "https://huggingface.co/birder-project/iformer_s_arabian-peninsula/resolve/main",
        "description": "iFormer small model trained on the arabian-peninsula dataset",
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 79.5,
                "sha256": "28ccab8f908e2d8129bb968d2142b5f7d851f0f1e9d522c7b76512d223a944de",
            }
        },
        "net": {"network": "iformer_s", "tag": "arabian-peninsula"},
    },
)



"""
DINO v1 (self-DIstillation with NO labels), adapted from
https://github.com/facebookresearch/dino/blob/main/main_dino.py

Paper "Emerging Properties in Self-Supervised Vision Transformers", https://arxiv.org/abs/2104.14294
"""

# Reference license: Apache-2.0

from typing import Any
from typing import Optional

import torch
import torch.distributed as dist
import torch.nn.functional as F
from torch import nn

from birder.common import training_utils
from birder.net.base import BaseNet
from birder.net.ssl.base import SSLBaseNet


class DINOLoss(nn.Module):
    def __init__(
        self,
        out_dim: int,
        num_crops: int,
        warmup_teacher_temp: float,
        teacher_temp: float,
        warmup_teacher_temp_epochs: int,
        num_epochs: int,
        student_temp: float,
        center_momentum: float,
    ) -> None:
        super().__init__()
        self.student_temp = student_temp
        self.center_momentum = center_momentum
        self.num_crops = num_crops
        self.center = nn.Buffer(torch.zeros(1, out_dim))

        # We apply a warm up for the teacher temperature because
        # a too high temperature makes the training instable at the beginning
        self.teacher_temp_schedule = torch.concat(
            (
                torch.linspace(warmup_teacher_temp, teacher_temp, warmup_teacher_temp_epochs),
                torch.ones(num_epochs - warmup_teacher_temp_epochs) * teacher_temp,
            ),
            dim=0,
        )

    def forward(self, student_output: torch.Tensor, teacher_output: torch.Tensor, epoch: int) -> torch.Tensor:
        student_out = student_output / self.student_temp
        student_out = student_out.chunk(self.num_crops)

        # Teacher centering and sharpening
        temp = self.teacher_temp_schedule[epoch]
        teacher_out = F.softmax((teacher_output - self.center) / temp, dim=-1)
        teacher_out = teacher_out.detach().chunk(2)

        total_loss = 0.0
        n_loss_terms = 0
        for iq, q in enumerate(teacher_out):
            for v in range(len(student_out)):  # pylint: disable=consider-using-enumerate
                if v == iq:
                    # Skip cases where student and teacher operate on the same view
                    continue

                loss = torch.sum(-q * F.log_softmax(student_out[v], dim=-1), dim=-1)
                total_loss += loss.mean()
                n_loss_terms += 1

        total_loss /= n_loss_terms
        self.update_center(teacher_output)

        return total_loss

    @torch.no_grad()  # type: ignore[misc]
    def update_center(self, teacher_output: torch.Tensor) -> None:
        batch_center = torch.sum(teacher_output, dim=0, keepdim=True)
        if training_utils.is_dist_available_and_initialized() is False:
            batch_center = batch_center / len(teacher_output)
        else:
            dist.all_reduce(batch_center)
            batch_center = batch_center / (len(teacher_output) * dist.get_world_size())

        # EMA update
        self.center = self.center * self.center_momentum + batch_center * (1 - self.center_momentum)


class DINOHead(nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        use_bn: bool,
        norm_last_layer: bool,
        num_layers: int,
        hidden_dim: int,
        bottleneck_dim: int,
    ) -> None:
        super().__init__()
        if num_layers == 1:
            self.mlp = nn.Linear(in_dim, bottleneck_dim)
        else:
            layers = [nn.Linear(in_dim, hidden_dim)]
            if use_bn is True:
                layers.append(nn.BatchNorm1d(hidden_dim))

            layers.append(nn.GELU())
            for _ in range(num_layers - 2):
                layers.append(nn.Linear(hidden_dim, hidden_dim))
                if use_bn is True:
                    layers.append(nn.BatchNorm1d(hidden_dim))

                layers.append(nn.GELU())

            layers.append(nn.Linear(hidden_dim, bottleneck_dim))
            self.mlp = nn.Sequential(*layers)

        self.last_layer = nn.utils.parametrizations.weight_norm(nn.Linear(bottleneck_dim, out_dim, bias=False))
        self.last_layer.parametrizations.weight.original0.data.fill_(1)
        if norm_last_layer is True:
            self.last_layer.parametrizations.weight.original0.requires_grad = False

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.mlp(x)
        x = F.normalize(x, dim=-1, p=2)
        x = self.last_layer(x)

        return x

    def cancel_last_layer_gradients(self) -> None:
        self.last_layer.zero_grad()


# pylint: disable=invalid-name
class DINO_v1(SSLBaseNet):
    def __init__(
        self,
        backbone: BaseNet,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(backbone, config=config, size=size)
        assert self.config is not None, "must set config"

        out_dim: int = self.config["out_dim"]
        use_bn: bool = self.config["use_bn"]
        norm_last_layer: bool = self.config["norm_last_layer"]
        num_layers: int = self.config["num_layers"]
        hidden_dim: int = self.config["hidden_dim"]
        bottleneck_dim: int = self.config["bottleneck_dim"]

        self.head = DINOHead(
            self.backbone.embedding_size,
            out_dim,
            use_bn=use_bn,
            norm_last_layer=norm_last_layer,
            num_layers=num_layers,
            hidden_dim=hidden_dim,
            bottleneck_dim=bottleneck_dim,
        )

    def forward(self, xs: list[torch.Tensor]) -> torch.Tensor:  # pylint: disable=arguments-renamed
        idx_crops = torch.cumsum(
            torch.unique_consecutive(
                torch.tensor([x.size(-1) for x in xs]),
                return_counts=True,
            )[1],
            0,
        )

        start_idx = 0
        output = torch.empty(0).to(xs[0].device)
        for end_idx in idx_crops:
            out = self.backbone.embedding(torch.concat(xs[start_idx:end_idx], dim=0))

            # Accumulate outputs
            output = torch.concat((output, out), dim=0)
            start_idx = end_idx

        return self.head(output)



"""
Adapted from https://github.com/jacobgil/pytorch-grad-cam

Paper "Striving for Simplicity: The All Convolutional Net", https://arxiv.org/abs/1412.6806
"""

# Reference license: MIT

from pathlib import Path
from typing import Any
from typing import Optional

import numpy as np
import numpy.typing as npt
import torch
from PIL import Image
from torch import nn
from torch.autograd import Function

from birder.introspection.base import InterpretabilityResult
from birder.introspection.base import Interpreter


def _deprocess_image(img: npt.NDArray[np.float32]) -> npt.NDArray[np.uint8]:
    """
    See https://github.com/jacobgil/keras-grad-cam/blob/master/grad-cam.py#L65
    """

    img = img - np.mean(img)
    img = img / (np.std(img) + 1e-5)
    img = img * 0.1
    img = img + 0.5
    img = np.clip(img, 0, 1)

    return np.array(img * 255).astype(np.uint8)


# pylint: disable=protected-access
def _replace_all_layer_type_recursive(model: nn.Module, old_layer_type: nn.Module, new_layer: nn.Module) -> None:
    for name, layer in model._modules.items():
        if isinstance(layer, old_layer_type):
            model._modules[name] = new_layer

        _replace_all_layer_type_recursive(layer, old_layer_type, new_layer)


# pylint: disable=abstract-method,arguments-differ
class GuidedBackpropReLU(Function):
    @staticmethod
    def forward(ctx: Any, input_img: torch.Tensor) -> torch.Tensor:
        positive_mask = (input_img > 0).type_as(input_img)
        output = torch.addcmul(torch.zeros(input_img.size()).type_as(input_img), input_img, positive_mask)
        ctx.save_for_backward(input_img, output)

        return output

    @staticmethod
    def backward(ctx: Any, grad_output: torch.Tensor) -> torch.Tensor:
        (input_img, _output) = ctx.saved_tensors
        grad_input = None

        positive_mask_1 = (input_img > 0).type_as(grad_output)
        positive_mask_2 = (grad_output > 0).type_as(grad_output)
        grad_input = torch.addcmul(
            torch.zeros(input_img.size()).type_as(input_img),
            torch.addcmul(torch.zeros(input_img.size()).type_as(input_img), grad_output, positive_mask_1),
            positive_mask_2,
        )

        return grad_input


# pylint: disable=abstract-method,arguments-differ
class GuidedBackpropSwish(Function):
    @staticmethod
    def forward(ctx: Any, input_img: torch.Tensor) -> torch.Tensor:
        result = input_img * torch.sigmoid(input_img)
        ctx.save_for_backward(input_img)

        return result

    @staticmethod
    def backward(ctx: Any, grad_output: torch.Tensor) -> torch.Tensor:
        i = ctx.saved_tensors[0]
        sigmoid_i = torch.sigmoid(i)
        positive_mask_1 = (i > 0).type_as(grad_output)
        positive_mask_2 = (grad_output > 0).type_as(grad_output)
        grad_input = grad_output * (sigmoid_i * (1 + i * (1 - sigmoid_i))) * positive_mask_1 * positive_mask_2

        return grad_input


class GuidedBackpropReLUAsModule(nn.Module):
    def forward(self, input_img: torch.Tensor) -> Any:
        return GuidedBackpropReLU.apply(input_img)


class GuidedBackpropSwishAsModule(nn.Module):
    def forward(self, input_img: torch.Tensor) -> Any:
        return GuidedBackpropSwish.apply(input_img)


class GuidedBackpropGeLUAsModule(nn.Module):
    def forward(self, input_img: torch.Tensor) -> Any:
        return GuidedBackpropSwish.apply(input_img)


class GuidedBackpropHardswishAsModule(nn.Module):
    def forward(self, input_img: torch.Tensor) -> Any:
        return GuidedBackpropSwish.apply(input_img)


class GuidedBackpropModel:
    def __init__(self, model: nn.Module) -> None:
        self.model = model
        self.model.eval()

    def forward(self, input_img: torch.Tensor) -> torch.Tensor:
        return self.model(input_img)

    def __call__(self, input_img: torch.Tensor, target_category: Optional[int] = None) -> npt.NDArray[np.float32]:
        _replace_all_layer_type_recursive(self.model, nn.ReLU, GuidedBackpropReLUAsModule())
        _replace_all_layer_type_recursive(self.model, nn.GELU, GuidedBackpropGeLUAsModule())
        _replace_all_layer_type_recursive(self.model, nn.SiLU, GuidedBackpropSwishAsModule())
        _replace_all_layer_type_recursive(self.model, nn.Hardswish, GuidedBackpropHardswishAsModule())

        input_img = input_img.requires_grad_(True)
        output = self.forward(input_img)

        if target_category is None:
            target_category = np.argmax(output.cpu().data.numpy()).item()

        loss = output[0, target_category]
        loss.backward(retain_graph=True)

        output_grad = input_img.grad.cpu().data.numpy()
        output_grad = output_grad[0, :, :, :]
        output_grad = output_grad.transpose((1, 2, 0))

        _replace_all_layer_type_recursive(self.model, GuidedBackpropHardswishAsModule, nn.Hardswish())
        _replace_all_layer_type_recursive(self.model, GuidedBackpropSwishAsModule, nn.SiLU())
        _replace_all_layer_type_recursive(self.model, GuidedBackpropGeLUAsModule, nn.GELU())
        _replace_all_layer_type_recursive(self.model, GuidedBackpropReLUAsModule, nn.ReLU())

        return output_grad  # type: ignore[no-any-return]


class GuidedBackpropInterpreter(Interpreter):
    def interpret(self, image: str | Path | Image.Image, target_class: Optional[int] = None) -> InterpretabilityResult:
        (input_tensor, rgb_img) = self._preprocess_image(image)

        guided_bp = GuidedBackpropModel(self.model)
        bp_img = guided_bp(input_tensor, target_category=target_class)

        return InterpretabilityResult(rgb_img, _deprocess_image(bp_img * rgb_img), raw_output=bp_img)



"""
SqueezeNext 23v5 version.

Paper "SqueezeNext: Hardware-Aware Neural Network Design",  https://arxiv.org/abs/1803.10615
"""

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class SqnxtUnit(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, stride: int) -> None:
        super().__init__()
        if stride == 2:
            reduction = 1
            self.identity = Conv2dNormActivation(
                in_channels,
                out_channels,
                kernel_size=(1, 1),
                stride=(stride, stride),
                padding=(0, 0),
                bias=False,
            )

        elif in_channels > out_channels:
            reduction = 4
            self.identity = Conv2dNormActivation(
                in_channels,
                out_channels,
                kernel_size=(1, 1),
                stride=(stride, stride),
                padding=(0, 0),
                bias=False,
            )

        else:
            reduction = 2
            self.identity = nn.Identity()

        self.block = nn.Sequential(
            Conv2dNormActivation(
                in_channels,
                in_channels // reduction,
                kernel_size=(1, 1),
                stride=(stride, stride),
                padding=(0, 0),
                bias=False,
            ),
            Conv2dNormActivation(
                in_channels // reduction,
                in_channels // (2 * reduction),
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
            ),
            Conv2dNormActivation(
                in_channels // (2 * reduction),
                in_channels // reduction,
                kernel_size=(1, 3),
                stride=(1, 1),
                padding=(0, 1),
                bias=False,
            ),
            Conv2dNormActivation(
                in_channels // reduction,
                in_channels // reduction,
                kernel_size=(3, 1),
                stride=(1, 1),
                padding=(1, 0),
                bias=False,
            ),
            Conv2dNormActivation(
                in_channels // reduction,
                out_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
            ),
        )
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = self.identity(x)
        x = self.block(x)
        x = x + identity
        x = self.relu(x)

        return x


class SqueezeNext(DetectorBackbone):
    default_size = (227, 227)

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        width_scale: float = self.config["width_scale"]

        channels_per_layers = [32, 64, 128, 256]
        layers_per_stage = [2, 4, 14, 1]

        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels,
                int(64 * width_scale),
                kernel_size=(7, 7),
                stride=(2, 2),
                padding=(1, 1),
                bias=False,
            ),
            nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(0, 0), ceil_mode=True),
        )

        in_channels = int(64 * width_scale)
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i, lps in enumerate(layers_per_stage):
            layers = []
            for j in range(lps):
                if j == 0 and i != 0:
                    stride = 2
                else:
                    stride = 1

                out_channels = int(channels_per_layers[i] * width_scale)
                layers.append(SqnxtUnit(in_channels, out_channels, stride))
                in_channels = out_channels

            stages[f"stage{i+1}"] = nn.Sequential(*layers)
            return_channels.append(out_channels)

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.Conv2d(
                in_channels,
                int(128 * width_scale),
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
            ),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = int(128 * width_scale)
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config("squeezenext_0_5", SqueezeNext, config={"width_scale": 0.5})
registry.register_model_config("squeezenext_1_0", SqueezeNext, config={"width_scale": 1.0})
registry.register_model_config("squeezenext_1_5", SqueezeNext, config={"width_scale": 1.5})
registry.register_model_config("squeezenext_2_0", SqueezeNext, config={"width_scale": 2.0})

registry.register_weights(
    "squeezenext_1_0_il-common",
    {
        "description": "SqueezeNext v2 1.0x output channels model trained on the il-common dataset",
        "resolution": (259, 259),
        "formats": {
            "pt": {
                "file_size": 3.5,
                "sha256": "da01d1cd05c71b80b5e4e6ca66400f64fa3f6179d0e90834c4f6942c8095557a",
            }
        },
        "net": {"network": "squeezenext_1_0", "tag": "il-common"},
    },
)



"""
Paper "ImageNet Classification with Deep Convolutional Neural Networks",
https://papers.nips.cc/paper/2012/hash/c399862d3b9d6b76c8436e924a68c45b-Abstract.html
"""

from typing import Any
from typing import Optional

import torch
from torch import nn

from birder.net.base import BaseNet


class AlexNet(BaseNet):
    auto_register = True

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is None, "config not supported"

        self.body = nn.Sequential(
            nn.Conv2d(self.input_channels, 64, kernel_size=(11, 11), stride=(4, 4), padding=(2, 2), bias=True),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(0, 0)),
            nn.Conv2d(64, 192, kernel_size=(5, 5), stride=(1, 1), padding=(2, 2), bias=True),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(0, 0)),
            nn.Conv2d(192, 384, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=True),
            nn.ReLU(inplace=True),
            nn.Conv2d(384, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=True),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=True),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(0, 0)),
            nn.AdaptiveAvgPool2d(output_size=(6, 6)),
        )
        self.features = nn.Sequential(
            nn.Flatten(1),
            nn.Linear(256 * 6 * 6, 4096),
            nn.ReLU(inplace=True),
            nn.Dropout(p=0.5),
            nn.Linear(4096, 4096),
            nn.ReLU(inplace=True),
            nn.Dropout(p=0.5),
        )
        self.embedding_size = 4096
        self.classifier = self.create_classifier()

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)



import logging
import unittest

import torch
import torch.nn.functional as F

from birder.common import masking
from birder.model_registry import registry
from birder.net.ssl import barlow_twins
from birder.net.ssl import byol
from birder.net.ssl import capi
from birder.net.ssl import data2vec
from birder.net.ssl import data2vec2
from birder.net.ssl import dino_v1
from birder.net.ssl import dino_v2
from birder.net.ssl import i_jepa
from birder.net.ssl import ibot
from birder.net.ssl import mmcr
from birder.net.ssl import simclr
from birder.net.ssl import sscd
from birder.net.ssl import vicreg

logging.disable(logging.CRITICAL)


class TestNetSSL(unittest.TestCase):
    def test_barlow_twins(self) -> None:
        batch_size = 4
        backbone = registry.net_factory("resnet_v1_50", 3, 0)
        net = barlow_twins.BarlowTwins(backbone, config={"projector_sizes": [512, 512, 512], "off_lambda": 0.005})

        # Test network
        out = net(torch.rand((batch_size, 3, 128, 128)), torch.rand((batch_size, 3, 128, 128)))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.ndim, 0)

    def test_byol(self) -> None:
        batch_size = 2
        backbone = registry.net_factory("resnet_v1_18", 3, 0)
        net = byol.BYOL(backbone, config={"projection_size": 64, "projection_hidden_size": 128})

        # Test network
        out = net(torch.rand((batch_size, 3, 96, 96)))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.ndim, 0)

    def test_capi(self) -> None:
        batch_size = 4
        size = (192, 192)
        num_clusters = 320
        backbone = registry.net_factory("vit_s16", 3, 0, size=size)
        input_size = (size[0] // backbone.max_stride, size[1] // backbone.max_stride)
        seq_len = input_size[0] * input_size[1]
        n_masked = int(seq_len * 0.65)
        prediction_subsampling = 0.05

        teacher = capi.CAPITeacher(
            backbone,
            config={"num_clusters": num_clusters, "bias": True, "n_sk_iter": 3, "target_temp": 0.06, "pred_temp": 0.12},
        )
        student = capi.CAPIStudent(
            backbone, config={"decoder_layers": 4, "decoder_dim": 256, "num_clusters": num_clusters}
        )
        mask_generator = masking.InverseRollBlockMasking(input_size, n_masked)

        masks = mask_generator(batch_size)
        ids_keep = masking.get_ids_keep(masks)
        n_predict = int(n_masked * prediction_subsampling)
        predict_indices = masking.get_random_masked_indices(masks, n_predict)
        masks = masking.mask_from_indices(predict_indices, seq_len)

        x = torch.rand(batch_size, 3, *size)
        # all_ids = torch.arange(input_size[0] * input_size[1]).unsqueeze(0).repeat(batch_size, 1)

        #
        # Simulate a full step
        #

        # Teacher
        (selected_assignments, clustering_loss) = teacher(x, None, predict_indices)
        self.assertFalse(torch.isnan(clustering_loss).any())
        self.assertFalse(torch.isnan(selected_assignments).any())
        self.assertEqual(selected_assignments.size(), (batch_size * n_predict, num_clusters))

        # Student
        pred = student(x, ids_keep, predict_indices)
        self.assertEqual(pred.size(), (masks.count_nonzero().item(), num_clusters))
        self.assertFalse(torch.isnan(pred).any())

        # Loss
        loss = -torch.sum(selected_assignments * F.log_softmax(pred / 0.12, dim=-1), dim=-1)
        self.assertFalse(torch.isnan(loss.sum() / len(loss)).any())
        self.assertEqual(loss.sum().ndim, 0)

        # Test with Hiera backbone
        backbone = registry.net_factory("hiera_tiny", 3, 0, size=size)
        input_size = (size[0] // backbone.max_stride, size[1] // backbone.max_stride)
        seq_len = input_size[0] * input_size[1]
        n_masked = int(seq_len * 0.65)
        prediction_subsampling = 0.05

        teacher = capi.CAPITeacher(
            backbone,
            config={"num_clusters": num_clusters, "bias": True, "n_sk_iter": 3, "target_temp": 0.06, "pred_temp": 0.12},
        )
        student = capi.CAPIStudent(
            backbone, config={"decoder_layers": 4, "decoder_dim": 256, "num_clusters": num_clusters}
        )
        mask_generator = masking.InverseRollBlockMasking(input_size, n_masked)

        masks = mask_generator(batch_size)
        ids_keep = masking.get_ids_keep(masks)
        n_predict = int(n_masked * prediction_subsampling)
        predict_indices = masking.get_random_masked_indices(masks, n_predict)
        masks = masking.mask_from_indices(predict_indices, seq_len)

        x = torch.rand(batch_size, 3, *size)

        # Teacher
        (selected_assignments, clustering_loss) = teacher(x, None, predict_indices)
        self.assertFalse(torch.isnan(clustering_loss).any())
        self.assertFalse(torch.isnan(selected_assignments).any())
        self.assertEqual(selected_assignments.size(), (batch_size * n_predict, num_clusters))

        # Student
        pred = student(x, ids_keep, predict_indices)
        self.assertEqual(pred.size(), (masks.count_nonzero().item(), num_clusters))
        self.assertFalse(torch.isnan(pred).any())

    def test_data2vec(self) -> None:
        batch_size = 2
        backbone = registry.net_factory("vit_t16", 3, 0, size=(96, 96))
        net = data2vec.Data2Vec(
            backbone, config={"normalize_targets": True, "average_top_k_layers": 6, "loss_beta": 2.0}
        )

        mask_generator = masking.BlockMasking((96 // backbone.stem_stride, 96 // backbone.stem_stride), 1, 3, 0.66, 1.5)
        masks = mask_generator(batch_size)

        # Test network
        out = net(torch.rand((batch_size, 3, 96, 96)), masks)
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.ndim, 0)

    def test_data2vec2(self) -> None:
        batch_size = 2
        backbone = registry.net_factory("vit_t16", 3, 0, size=(128, 128))
        net = data2vec2.Data2Vec2(
            backbone,
            config={
                "average_top_k_layers": 6,
                "decoder_dim": 128,
                "decoder_kernel_size": 3,
                "decoder_layers": 4,
                "clone_batch": 3,
                "cls_loss_weight": 0.1,
            },
        )

        mask_generator = masking.InverseRollBlockMasking(
            (128 // backbone.stem_stride, 128 // backbone.stem_stride),
            num_masking_patches=32,
            min_aspect=0.66,
            max_aspect=1.5,
        )
        masks = mask_generator(batch_size * 3)

        # Test network
        out = net(torch.rand((batch_size, 3, 128, 128)), masks)
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.ndim, 0)

    def test_dino_v1(self) -> None:
        batch_size = 4
        backbone = registry.net_factory("resnet_v2_18", 3, 0)
        net = dino_v1.DINO_v1(
            backbone,
            config={
                "out_dim": 128,
                "use_bn": True,
                "norm_last_layer": True,
                "num_layers": 3,
                "hidden_dim": 512,
                "bottleneck_dim": 256,
            },
        )

        # Test network
        out = net(
            [
                torch.rand((batch_size, 3, 128, 128)),
                torch.rand((batch_size, 3, 128, 128)),
                torch.rand((batch_size, 3, 96, 96)),
                torch.rand((batch_size, 3, 96, 96)),
                torch.rand((batch_size, 3, 96, 96)),
                torch.rand((batch_size, 3, 96, 96)),
            ]
        )
        self.assertFalse(torch.isnan(out).any())

        teacher_out = net([torch.rand((batch_size, 3, 128, 128)), torch.rand((batch_size, 3, 128, 128))])
        dino_loss = dino_v1.DINOLoss(
            128,
            6,
            warmup_teacher_temp=0.2,
            teacher_temp=0.8,
            warmup_teacher_temp_epochs=10,
            num_epochs=100,
            student_temp=0.5,
            center_momentum=0.99,
        )
        loss = dino_loss(out, teacher_out, epoch=2)
        self.assertFalse(torch.isnan(loss).any())
        self.assertEqual(loss.ndim, 0)

    # pylint: disable=too-many-locals
    def test_dino_v2(self) -> None:
        batch_size = 4
        size = (192, 192)
        backbone = registry.net_factory("vit_s16", 3, 0, size=size)
        backbone.set_dynamic_size()

        # Without iBOT head
        teacher = dino_v2.DINOv2Teacher(
            backbone,
            config={
                "dino_out_dim": 4096,
                "use_bn": False,
                "num_layers": 3,
                "hidden_dim": 2048,
                "head_bottleneck_dim": 256,
                "ibot_separate_head": False,
            },
        )
        student = dino_v2.DINOv2Student(
            backbone,
            config={
                "dino_out_dim": 4096,
                "use_bn": False,
                "num_layers": 3,
                "hidden_dim": 2048,
                "head_bottleneck_dim": 256,
                "ibot_separate_head": False,
            },
        )

        x = torch.rand(batch_size * 2, 3, *size)
        local_x = torch.rand(batch_size * 4, 3, size[0] // 2, size[1] // 2)
        mask_generator = masking.BlockMasking(
            (size[0] // backbone.stem_stride, size[1] // backbone.stem_stride), 1, 3, 0.66, 1.5
        )
        masks = mask_generator(batch_size * 2)  # Not quite the same, but close enough
        mask_indices_list = masks.flatten().nonzero().flatten()
        upper_bound = len(mask_indices_list) + 16

        with torch.no_grad():
            (teacher_embedding_after_head, teacher_masked_patch_tokens_after_head) = teacher(
                x, 2, upper_bound=upper_bound, mask_indices_list=mask_indices_list
            )

        (
            student_global_embedding,
            student_global_embedding_after_head,
            student_local_embedding_after_head,
            student_global_masked_patch_tokens_after_head,
        ) = student(x, local_x, masks, upper_bound, mask_indices_list)

        self.assertFalse(torch.isnan(teacher_embedding_after_head).any())
        self.assertEqual(teacher_embedding_after_head.size(), (batch_size * 2, 4096))
        self.assertFalse(torch.isnan(teacher_masked_patch_tokens_after_head).any())
        self.assertEqual(teacher_masked_patch_tokens_after_head.size(), (len(mask_indices_list), 4096))

        self.assertFalse(torch.isnan(student_global_embedding).any())
        self.assertEqual(student_global_embedding.size(), (batch_size * 2, backbone.embedding_size))
        self.assertFalse(torch.isnan(student_global_embedding_after_head).any())
        self.assertEqual(student_global_embedding_after_head.size(), (batch_size * 2, 4096))
        self.assertFalse(torch.isnan(student_local_embedding_after_head).any())
        self.assertEqual(student_local_embedding_after_head.size(), (batch_size * 4, 4096))
        self.assertFalse(torch.isnan(student_global_masked_patch_tokens_after_head).any())
        self.assertEqual(student_global_masked_patch_tokens_after_head.size(), (len(mask_indices_list), 4096))

        # With iBOT head
        teacher = dino_v2.DINOv2Teacher(
            backbone,
            config={
                "dino_out_dim": 4096,
                "use_bn": False,
                "num_layers": 3,
                "hidden_dim": 2048,
                "head_bottleneck_dim": 256,
                "ibot_separate_head": True,
                "ibot_out_dim": 4096,
            },
        )
        student = dino_v2.DINOv2Student(
            backbone,
            config={
                "dino_out_dim": 4096,
                "use_bn": False,
                "num_layers": 3,
                "hidden_dim": 2048,
                "head_bottleneck_dim": 256,
                "ibot_separate_head": True,
                "ibot_out_dim": 4096,
            },
        )

        x = torch.rand(batch_size * 2, 3, *size)
        local_x = torch.rand(batch_size * 4, 3, size[0] // 2, size[1] // 2)
        mask_generator = masking.BlockMasking(
            (size[0] // backbone.stem_stride, size[1] // backbone.stem_stride), 1, 3, 0.66, 1.5
        )
        masks = mask_generator(batch_size * 2)  # Not quite the same, but close enough
        masks = masks.to(torch.long)
        mask_indices_list = masks.flatten().nonzero().flatten()
        upper_bound = len(mask_indices_list) + 16

        with torch.no_grad():
            (teacher_embedding_after_head, teacher_masked_patch_tokens_after_head) = teacher(
                x, 2, upper_bound=upper_bound, mask_indices_list=mask_indices_list
            )

        (
            student_global_embedding,
            student_global_embedding_after_head,
            student_local_embedding_after_head,
            student_global_masked_patch_tokens_after_head,
        ) = student(x, local_x, masks, upper_bound, mask_indices_list)

        self.assertFalse(torch.isnan(teacher_embedding_after_head).any())
        self.assertEqual(teacher_embedding_after_head.size(), (batch_size * 2, 4096))
        self.assertFalse(torch.isnan(teacher_masked_patch_tokens_after_head).any())
        self.assertEqual(teacher_masked_patch_tokens_after_head.size(), (len(mask_indices_list), 4096))

        self.assertFalse(torch.isnan(student_global_embedding).any())
        self.assertEqual(student_global_embedding.size(), (batch_size * 2, backbone.embedding_size))
        self.assertFalse(torch.isnan(student_global_embedding_after_head).any())
        self.assertEqual(student_global_embedding_after_head.size(), (batch_size * 2, 4096))
        self.assertFalse(torch.isnan(student_local_embedding_after_head).any())
        self.assertEqual(student_local_embedding_after_head.size(), (batch_size * 4, 4096))
        self.assertFalse(torch.isnan(student_global_masked_patch_tokens_after_head).any())
        self.assertEqual(student_global_masked_patch_tokens_after_head.size(), (len(mask_indices_list), 4096))

        # Loss
        dino_loss = dino_v2.DINOLoss(4096, student_temp=0.1, center_momentum=0.9)
        koleo_loss = dino_v2.KoLeoLoss()
        ibot_patch_loss = dino_v2.iBOTPatchLoss(4096, student_temp=0.1, center_momentum=0.9)

        # Loss centering
        teacher_temp = 0.04
        n_masked_patches = len(mask_indices_list)

        teacher_dino_softmax_centered_list = dino_loss.softmax_center_teacher(
            teacher_embedding_after_head, teacher_temp=teacher_temp
        ).view(2, -1, *teacher_embedding_after_head.shape[1:])
        dino_loss.update_center(teacher_embedding_after_head)

        teacher_masked_patch_tokens_after_head_d = teacher_masked_patch_tokens_after_head.unsqueeze(0)
        masked_teacher_ibot_softmax_centered = ibot_patch_loss.softmax_center_teacher(
            teacher_masked_patch_tokens_after_head_d[:, :n_masked_patches], teacher_temp=teacher_temp
        )
        masked_teacher_ibot_softmax_centered = masked_teacher_ibot_softmax_centered.squeeze(0)
        ibot_patch_loss.update_center(teacher_masked_patch_tokens_after_head_d[:n_masked_patches])

        self.assertFalse(torch.isnan(teacher_dino_softmax_centered_list).any())
        self.assertEqual(teacher_dino_softmax_centered_list.size(), (2, batch_size, 4096))
        self.assertFalse(torch.isnan(masked_teacher_ibot_softmax_centered).any())
        self.assertEqual(masked_teacher_ibot_softmax_centered.size(), (len(mask_indices_list), 4096))

        # Loss centering - Sinkhorn Knopp
        teacher_dino_softmax_centered_list = dino_loss.sinkhorn_knopp_teacher(
            teacher_embedding_after_head, teacher_temp=teacher_temp
        ).view(2, -1, *teacher_embedding_after_head.shape[1:])

        masked_teacher_ibot_softmax_centered = ibot_patch_loss.sinkhorn_knopp_teacher(
            teacher_masked_patch_tokens_after_head,
            teacher_temp=teacher_temp,
            n_masked_patches_tensor=torch.full((1,), fill_value=mask_indices_list.size(0), dtype=torch.long),
        )

        self.assertFalse(torch.isnan(teacher_dino_softmax_centered_list).any())
        self.assertEqual(teacher_dino_softmax_centered_list.size(), (2, batch_size, 4096))
        self.assertFalse(torch.isnan(masked_teacher_ibot_softmax_centered).any())
        self.assertEqual(masked_teacher_ibot_softmax_centered.size(), (len(mask_indices_list), 4096))

        # DINO loss
        loss_dino = dino_loss(student_local_embedding_after_head.chunk(4), teacher_dino_softmax_centered_list)
        self.assertFalse(torch.isnan(loss_dino).any())
        self.assertEqual(loss_dino.ndim, 0)

        # KoLeo loss
        loss_koleo = sum(koleo_loss(p) for p in student_global_embedding.chunk(2))
        self.assertFalse(torch.isnan(loss_koleo).any())
        self.assertEqual(loss_koleo.ndim, 0)

        # iBOT loss
        masks_weight = (1 / masks.sum(-1).clamp(min=1.0)).unsqueeze(-1).expand_as(masks)[masks.bool()]
        loss_ibot_patch = ibot_patch_loss.forward_masked(
            student_global_masked_patch_tokens_after_head,
            masked_teacher_ibot_softmax_centered,
            student_masks_flat=masks,
            masks_weight=masks_weight,
            n_masked_patches=n_masked_patches,
        )
        self.assertFalse(torch.isnan(loss_ibot_patch).any())
        self.assertEqual(loss_ibot_patch.ndim, 0)

    def test_i_jepa(self) -> None:
        batch_size = 4
        size = (192, 192)
        backbone = registry.net_factory("vit_s16", 3, 0, size=size)
        input_size = (size[0] // backbone.stem_stride, size[1] // backbone.stem_stride)
        predictor = i_jepa.VisionTransformerPredictor(
            input_size,
            backbone.embedding_size,
            128,
            512,
            8,
            2,
            0.0,
        )
        mask_generator = i_jepa.MultiBlockMasking(
            input_size,
            enc_mask_scale=(0.85, 1.0),
            pred_mask_scale=(0.15, 0.25),
            aspect_ratio=(0.75, 1.5),
            n_enc=2,
            n_pred=3,
            min_keep=10,
            allow_overlap=False,
        )
        masks = mask_generator(batch_size)
        enc_masks = masks[0]
        pred_masks = masks[1]

        x = torch.rand(batch_size, 3, *size)

        #
        # Simulate a full step
        #

        # Target encoder
        h = backbone.masked_encoding_omission(x)["tokens"]
        h = h[:, backbone.num_special_tokens :, :]  # Remove special tokens
        h = i_jepa.apply_masks(h, pred_masks)
        h = i_jepa.repeat_interleave_batch(h, batch_size, repeat=len(enc_masks))

        # Context
        z = torch.concat([backbone.masked_encoding_omission(x, enc_mask)["tokens"] for enc_mask in enc_masks], dim=0)
        z = z[:, backbone.num_special_tokens :, :]  # Remove special tokens
        z = predictor(z, enc_masks, pred_masks)

        loss = F.smooth_l1_loss(z, h)

        self.assertEqual(h.size(0), batch_size * len(enc_masks) * len(pred_masks))
        self.assertEqual(z.size(0), batch_size * len(enc_masks) * len(pred_masks))

        self.assertEqual(h.size(2), backbone.embedding_size)
        self.assertEqual(z.size(2), backbone.embedding_size)

        self.assertEqual(h.size(1), z.size(1))

        self.assertEqual(loss.ndim, 0)

    def test_ibot(self) -> None:
        batch_size = 4
        backbone = registry.net_factory("vit_b32", 3, 0)
        backbone.set_dynamic_size()
        net = ibot.iBOT(
            backbone,
            config={
                "out_dim": 128,
                "norm_last_layer": True,
                "num_layers": 3,
                "hidden_dim": 512,
                "bottleneck_dim": 256,
                "patch_out_dim": 192,
                "shared_head": False,
            },
        )

        # Test network
        images = [
            # Global
            torch.rand((batch_size, 3, 128, 128)),
            torch.rand((batch_size, 3, 128, 128)),
            # Local
            torch.rand((batch_size, 3, 96, 96)),
            torch.rand((batch_size, 3, 96, 96)),
            torch.rand((batch_size, 3, 96, 96)),
            torch.rand((batch_size, 3, 96, 96)),
        ]

        mask_generator = masking.BlockMasking(
            (128 // backbone.stem_stride, 128 // backbone.stem_stride), 1, 3, 0.66, 1.5
        )
        masks = mask_generator(batch_size * 2)

        (embedding_g, features_g) = net(torch.concat(images[:2], dim=0), masks=masks)
        self.assertFalse(torch.isnan(embedding_g).any())
        self.assertFalse(torch.isnan(features_g).any())
        self.assertEqual(features_g.size(), (batch_size * 2, (128 // 32) ** 2, 192))
        self.assertEqual(embedding_g.size(), (batch_size * 2, 128))

        (embedding_l, features_l) = net(torch.concat(images[2:], dim=0), masks=None)
        self.assertFalse(torch.isnan(embedding_l).any())
        self.assertFalse(torch.isnan(features_l).any())
        self.assertEqual(features_l.size(), (batch_size * 4, (96 // 32) ** 2, 192))
        self.assertEqual(embedding_l.size(), (batch_size * 4, 128))

        ibot_loss = ibot.iBOTLoss(
            128,
            192,
            num_global_crops=2,
            num_local_crops=4,
            warmup_teacher_temp=0.1,
            teacher_temp=0.9,
            warmup_teacher_temp2=0.2,
            teacher_temp2=0.99,
            warmup_teacher_temp_epochs=5,
            epochs=100,
            student_temp=0.5,
            center_momentum=0.98,
            center_momentum2=0.97,
            lambda1=0.2,
            lambda2=0.1,
            mim_start_epoch=1,
        )

        loss = ibot_loss.forward(
            embedding_g,
            features_g,
            torch.rand_like(embedding_g),
            torch.rand_like(features_g),
            student_local_embedding=embedding_l,
            student_mask=masks,
            epoch=2,
        )

        self.assertFalse(torch.isnan(loss["all"]).any())
        self.assertFalse(torch.isnan(loss["embedding"]).any())
        self.assertFalse(torch.isnan(loss["features"]).any())

        self.assertEqual(loss["all"].ndim, 0)
        self.assertEqual(loss["embedding"].ndim, 0)
        self.assertEqual(loss["features"].ndim, 0)

    def test_mmcr(self) -> None:
        batch_size = 4
        backbone = registry.net_factory("resnet_v1_50", 3, 0)
        net = mmcr.MMCR(backbone, config={"projector_dims": [512, 512, 512]})
        mmcr_loss = mmcr.MMCRMomentumLoss(0.0, 2)

        # Test network
        (out, out_m) = net(torch.rand((batch_size, 2, 3, 128, 128)))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.ndim, 2)
        self.assertFalse(torch.isnan(out_m).any())
        self.assertEqual(out_m.ndim, 2)

        # Test loss
        loss = mmcr_loss(out, out_m)
        self.assertFalse(torch.isnan(loss).any())
        self.assertEqual(loss.ndim, 0)

    def test_simclr(self) -> None:
        batch_size = 4
        backbone = registry.net_factory("resnet_v2_18", 3, 0)
        net = simclr.SimCLR(
            backbone,
            config={
                "projection_dim": 96,
                "projection_hidden_dim": 128,
                "temperature": 0.1,
            },
        )

        # Test network
        out = net(torch.rand((batch_size, 3, 128, 128)), torch.rand((batch_size, 3, 128, 128)))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.ndim, 0)

    def test_sscd(self) -> None:
        batch_size = 4
        backbone = registry.net_factory("resnet_v2_18", 3, 512)
        net = sscd.SSCD(backbone)
        out = net(torch.rand(batch_size, 3, 128, 128))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.ndim, 2)

    def test_vicreg(self) -> None:
        batch_size = 4
        backbone = registry.net_factory("resnet_v1_18", 3, 0)
        net = vicreg.VICReg(
            backbone,
            config={
                "mlp_dim": 128,
                "batch_size": batch_size,
                "sim_coeff": 0.1,
                "std_coeff": 0.1,
                "cov_coeff": 0.1,
            },
        )

        # Test network
        out = net(torch.rand((batch_size, 3, 128, 128)), torch.rand((batch_size, 3, 128, 128)))
        self.assertFalse(torch.isnan(out).any())
        self.assertEqual(out.ndim, 0)



"""
ResMLP, adapted from
https://github.com/facebookresearch/deit/blob/main/resmlp_models.py

Paper "ResMLP: Feedforward networks for image classification with data-efficient training",
https://arxiv.org/abs/2105.03404
"""

# Reference license: Apache-2.0

from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import Permute
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import BaseNet
from birder.net.cait import PatchEmbed


class Affine(nn.Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.alpha = nn.Parameter(torch.ones(dim))
        self.beta = nn.Parameter(torch.zeros(dim))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.alpha * x + self.beta


class LayerScaleMLP(nn.Module):
    def __init__(self, dim: int, num_patches: int, drop: float, drop_path: float, init_value: float) -> None:
        super().__init__()
        self.norm1 = Affine(dim)
        self.attn = nn.Linear(num_patches, num_patches)
        self.drop_path = StochasticDepth(drop_path, mode="row")
        self.norm2 = Affine(dim)
        self.mlp = MLP(dim, [int(dim * 4.0), dim], activation_layer=nn.GELU, dropout=drop)
        self.gamma_1 = nn.Parameter(init_value * torch.ones((dim)), requires_grad=True)
        self.gamma_2 = nn.Parameter(init_value * torch.ones((dim)), requires_grad=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.drop_path(self.gamma_1 * self.attn(self.norm1(x).transpose(1, 2)).transpose(1, 2))
        x = x + self.drop_path(self.gamma_2 * self.mlp(self.norm2(x)))

        return x


class ResMLP(BaseNet):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        init_value = 1e-4
        drop_rate = 0.0
        embed_dim: int = self.config["embed_dim"]
        depth: int = self.config["depth"]
        patch_size: tuple[int, int] = self.config["patch_size"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.patch_embed = PatchEmbed(self.input_channels, embed_dim, patch_size)
        num_patches = (self.size[0] // patch_size[0]) * (self.size[1] // patch_size[1])

        dpr = [drop_path_rate for _ in range(depth)]
        blocks = []
        for i in range(depth):
            blocks.append(
                LayerScaleMLP(
                    embed_dim, num_patches=num_patches, drop=drop_rate, drop_path=dpr[i], init_value=init_value
                )
            )

        blocks.append(Affine(embed_dim))

        self.body = nn.Sequential(*blocks)
        self.features = nn.Sequential(
            Permute([0, 2, 1]),
            nn.AdaptiveAvgPool1d(output_size=1),
            nn.Flatten(1),
        )
        self.embedding_size = embed_dim
        self.classifier = self.create_classifier()

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.patch_embed(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        raise RuntimeError("Model resizing not supported")


registry.register_model_config(
    "resmlp_12",
    ResMLP,
    config={"embed_dim": 384, "depth": 12, "patch_size": (16, 16), "init_value": 0.1, "drop_path_rate": 0.0},
)
registry.register_model_config(
    "resmlp_24",
    ResMLP,
    config={"embed_dim": 384, "depth": 24, "patch_size": (16, 16), "init_value": 1e-5, "drop_path_rate": 0.1},
)
registry.register_model_config(
    "resmlp_36",
    ResMLP,
    config={"embed_dim": 384, "depth": 36, "patch_size": (16, 16), "init_value": 1e-6, "drop_path_rate": 0.2},
)
registry.register_model_config(
    "resmlp_24_big",
    ResMLP,
    config={"embed_dim": 768, "depth": 24, "patch_size": (8, 8), "init_value": 1e-6, "drop_path_rate": 0.2},
)
